import unittest
from copy import copy, deepcopy

import numpy as np
import pinocchio as pin
from pinocchio.utils import eye, rand, zero


def ones(n):
    return np.ones([n, 1] if isinstance(n, int) else n)


class TestSE3Bindings(unittest.TestCase):
    def test_identity(self):
        transform = pin.SE3.Identity()
        self.assertTrue(np.allclose(zero(3), transform.translation))
        self.assertTrue(np.allclose(eye(3), transform.rotation))
        self.assertTrue(np.allclose(eye(4), transform.homogeneous))
        self.assertTrue(np.allclose(eye(6), transform.action))
        transform.setRandom()
        transform.setIdentity()
        self.assertTrue(np.allclose(eye(4), transform.homogeneous))

    def test_constructor(self):
        M = pin.SE3.Random()
        quat = pin.Quaternion(M.rotation)
        M_from_quat = pin.SE3(quat, M.translation)
        self.assertTrue(M_from_quat.isApprox(M))

    def test_get_translation(self):
        transform = pin.SE3.Identity()
        self.assertTrue(np.allclose(transform.translation, zero(3)))

    def test_get_rotation(self):
        transform = pin.SE3.Identity()
        self.assertTrue(np.allclose(transform.rotation, eye(3)))

    def test_set_translation(self):
        transform = pin.SE3.Identity()
        transform.translation = ones(3)
        self.assertFalse(np.allclose(transform.translation, zero(3)))
        self.assertTrue(np.allclose(transform.translation, ones(3)))

    def test_set_rotation(self):
        transform = pin.SE3.Identity()
        transform.rotation = zero([3, 3])
        self.assertFalse(np.allclose(transform.rotation, eye(3)))
        self.assertTrue(np.allclose(transform.rotation, zero([3, 3])))

    def test_homogeneous(self):
        amb = pin.SE3.Random()
        H = amb.homogeneous
        self.assertTrue(
            np.allclose(H[0:3, 0:3], amb.rotation)
        )  # top left 33 corner = rotation of amb
        self.assertTrue(
            np.allclose(H[0:3, 3], amb.translation)
        )  # top 3 of last column = translation of amb
        self.assertTrue(
            np.allclose(H[3, :], [0.0, 0.0, 0.0, 1.0])
        )  # last row = 0 0 0 1

        amb_from_H = pin.SE3(H)
        self.assertTrue(amb_from_H.isApprox(amb))

    def test_action_matrix(self):
        amb = pin.SE3.Random()
        aXb = amb.action
        self.assertTrue(
            np.allclose(aXb[:3, :3], amb.rotation)
        )  # top left 33 corner = rotation of amb
        self.assertTrue(
            np.allclose(aXb[3:, 3:], amb.rotation)
        )  # bottom right 33 corner = rotation of amb
        tblock = pin.skew(amb.translation).dot(amb.rotation)
        self.assertTrue(
            np.allclose(aXb[:3, 3:], tblock)
        )  # top right 33 corner = translation + rotation
        self.assertTrue(
            np.allclose(aXb[3:, :3], zero([3, 3]))
        )  # bottom left 33 corner = 0

    def test_inverse(self):
        amb = pin.SE3.Random()
        aMb = amb.homogeneous
        bma = amb.inverse()
        self.assertTrue(np.allclose(np.linalg.inv(aMb), bma.homogeneous))

    def test_internal_product_vs_homogeneous(self):
        amb = pin.SE3.Random()
        bmc = pin.SE3.Random()
        amc = amb * bmc
        cma = amc.inverse()

        aMb = amb.homogeneous
        bMc = bmc.homogeneous
        aMc = amc.homogeneous
        cMa = np.linalg.inv(aMc)

        self.assertTrue(np.allclose(aMb.dot(bMc), aMc))
        self.assertTrue(np.allclose(cma.homogeneous, cMa))

    def test_internal_product_vs_action(self):
        amb = pin.SE3.Random()
        bmc = pin.SE3.Random()
        amc = amb * bmc
        cma = amc.inverse()

        aXb = amb.action
        bXc = bmc.action
        aXc = amc.action
        cXa = np.linalg.inv(aXc)

        self.assertTrue(np.allclose(aXb.dot(bXc), aXc))
        self.assertTrue(np.allclose(cma.action, cXa))

    def test_point_action(self):
        amb = pin.SE3.Random()
        aMb = amb.homogeneous
        p_homogeneous = rand(4)
        p_homogeneous[3] = 1
        p = p_homogeneous[0:3].copy()

        # act
        self.assertTrue(np.allclose(amb.act(p), (aMb.dot(p_homogeneous))[0:3]))

        # actinv
        bMa = np.linalg.inv(aMb)
        bma = amb.inverse()
        self.assertTrue(np.allclose(bma.act(p), (bMa.dot(p_homogeneous))[0:3]))

    def test_member(self):
        M = pin.SE3.Random()
        trans = M.translation
        M.translation[2] = 1.0

        self.assertTrue(trans[2] == M.translation[2])

    def test_conversions(self):
        def compute(m):
            tq_vec = pin.SE3ToXYZQUAT(m)
            tq_tup = pin.SE3ToXYZQUATtuple(m)
            mm_vec = pin.XYZQUATToSE3(tq_vec)
            mm_tup = pin.XYZQUATToSE3(tq_tup)
            mm_lis = pin.XYZQUATToSE3(list(tq_tup))
            return tq_vec, tq_tup, mm_vec, mm_tup, mm_lis

        m = pin.SE3.Identity()
        tq_vec, tq_tup, mm_vec, mm_tup, mm_lis = compute(m)
        self.assertTrue(np.allclose(m.homogeneous, mm_tup.homogeneous))
        self.assertTrue(np.allclose(m.homogeneous, mm_vec.homogeneous))
        self.assertTrue(np.allclose(m.homogeneous, mm_lis.homogeneous))
        for i in range(6):
            self.assertTrue(tq_vec[i] == 0 and tq_tup[i] == 0)
        self.assertTrue(tq_vec[6] == 1 and tq_tup[6] == 1)

        m = pin.SE3.Random()
        tq_vec, tq_tup, mm_vec, mm_tup, mm_lis = compute(m)
        self.assertTrue(len(tq_vec) == 7)
        self.assertTrue(len(tq_tup) == 7)
        for a, b in zip(tq_vec, tq_tup):
            self.assertTrue(a == b)
        self.assertTrue(np.allclose(m.homogeneous, mm_tup.homogeneous))
        self.assertTrue(mm_vec == mm_tup)
        self.assertTrue(mm_vec == mm_lis)

        M = pin.SE3.Random()
        h = np.array(M)

        M_from_h = pin.SE3(h)
        self.assertTrue(M == M_from_h)

    def test_several_init(self):
        for _ in range(100000):
            r = pin.SE3.Random() * pin.SE3.Random()
            s = r.__str__()
            self.assertTrue(s != "")

    def test_copy(self):
        M = pin.SE3.Random()
        Mc = copy(M)
        Mdc = deepcopy(M)
        self.assertTrue(M == Mc)
        self.assertTrue(M == Mdc)

        Mc.setRandom()
        self.assertFalse(M == Mc)
        Mdc.setRandom()
        self.assertFalse(M == Mc)


if __name__ == "__main__":
    unittest.main()



import sys
import unittest
from pathlib import Path

sys.path.append(str(Path(__file__).parent.parent))
import casadi
import numpy as np
import pinocchio as pin
from casadi import SX
from pinocchio import casadi as cpin
from test_case import PinocchioTestCase as TestCase


class TestMainAlgos(TestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoidRandom()
        self.data = self.model.createData()

        self.cmodel = cpin.Model(self.model)
        self.cdata = self.cmodel.createData()

        qmax = np.full((self.model.nq, 1), np.pi)
        self.q = pin.randomConfiguration(self.model, -qmax, qmax)
        self.v = np.random.rand(self.model.nv)
        self.a = np.random.rand(self.model.nv)
        self.tau = np.random.rand(self.model.nv)

        self.cq = SX.sym("q", self.cmodel.nq, 1)
        self.cv = SX.sym("v", self.cmodel.nv, 1)
        self.ca = SX.sym("a", self.cmodel.nv, 1)
        self.ctau = SX.sym("tau", self.cmodel.nv, 1)

        self.fext = []
        self.cfext = []
        for i in range(self.model.njoints):
            self.fext.append(pin.Force.Random())
            fname = "f" + str(i)
            self.cfext.append(cpin.Force(SX.sym(fname, 6, 1)))

        self.fext[0].setZero()

    def test_crba(self):
        model = self.model
        data = self.data

        cmodel = self.cmodel
        cdata = self.cdata

        cM = cpin.crba(cmodel, cdata, self.cq)
        f_crba = casadi.Function("crba", [self.cq], [cM])

        arg_crba = [list(self.q)]
        M = f_crba.call(arg_crba)[0].full()

        M_ref = pin.crba(model, data, self.q)

        self.assertApprox(M, M_ref)

    def test_rnea(self):
        model = self.model
        data = self.data

        cmodel = self.cmodel
        cdata = self.cdata

        ctau = cpin.rnea(cmodel, cdata, self.cq, self.cv, self.ca)
        f_rnea = casadi.Function("rnea", [self.cq, self.cv, self.ca], [ctau])

        arg_rnea = [list(self.q), list(self.v), list(self.a)]
        tau = f_rnea.call(arg_rnea)[0].full()

        tau_ref = pin.rnea(model, data, self.q, self.v, self.a)

        self.assertApprox(tau, tau_ref)

        ctau_fext = cpin.rnea(cmodel, cdata, self.cq, self.cv, self.ca, self.cfext)
        carg_in = [self.cq, self.cv, self.ca]
        for f in self.cfext:
            carg_in += [f.vector]
        f_rnea_fext = casadi.Function("rnea_fext", carg_in, [ctau_fext])

        arg_rnea_fext = arg_rnea
        for f in self.fext:
            arg_rnea_fext += [f.vector]
        tau_fext = f_rnea_fext.call(arg_rnea_fext)[0].full()

        tau_fext_ref = pin.rnea(model, data, self.q, self.v, self.a, self.fext)

        self.assertApprox(tau_fext, tau_fext_ref)

    def test_aba(self):
        model = self.model
        data = self.data

        cmodel = self.cmodel
        cdata = self.cdata

        cddq = cpin.aba(cmodel, cdata, self.cq, self.cv, self.ctau)
        f_aba = casadi.Function("aba", [self.cq, self.cv, self.ctau], [cddq])

        arg_aba = [list(self.q), list(self.v), list(self.tau)]
        a = f_aba.call(arg_aba)[0].full()

        a_ref = pin.aba(model, data, self.q, self.v, self.tau)

        self.assertApprox(a, a_ref)

        cddq_fext = cpin.aba(cmodel, cdata, self.cq, self.cv, self.ctau, self.cfext)
        carg_in = [self.cq, self.cv, self.ctau]
        for f in self.cfext:
            carg_in += [f.vector]
        f_aba_fext = casadi.Function("aba_fext", carg_in, [cddq_fext])

        arg_aba_fext = arg_aba
        for f in self.fext:
            arg_aba_fext += [f.vector]
        a_fext = f_aba_fext.call(arg_aba_fext)[0].full()

        a_fext_ref = pin.aba(model, data, self.q, self.v, self.tau, self.fext)

        self.assertApprox(a_fext, a_fext_ref)

    def test_computeMinverse(self):
        model = self.model
        Minv = pin.computeMinverse(model, self.data, self.q)

        data2 = model.createData()
        M = pin.crba(model, data2, self.q)

        self.assertApprox(np.linalg.inv(M), Minv)


if __name__ == "__main__":
    unittest.main()



# This examples shows how to load and move a robot in panda3d_viewer.
# Note: this feature requires panda3d_viewer to be installed, this can be done using
# pip install panda3d_viewer
# ruff: noqa: E402


import sys
from pathlib import Path

import numpy as np

# Add path to the example-robot-data package from git submodule.
# If you have a proper install version, there is no need for this sys.path thing
path = Path(__file__).parent.parent / "models" / "example-robot-data" / "python"
sys.path.append(str(path))
from example_robot_data.robots_loader import TalosLoader
from panda3d_viewer import ViewerClosedError
from pinocchio.visualize.panda3d_visualizer import Panda3dVisualizer

# talos is a RobotWrapper object
talos = TalosLoader().robot
# Attach talos to the viewer scene
talos.setVisualizer(Panda3dVisualizer())
talos.initViewer()
talos.loadViewerModel(group_name="talos", color=(1, 1, 1, 1))
record_video = False


# Play a sample trajectory in a loop
def play_sample_trajectory():
    update_rate = 60
    cycle_time = 3
    traj = np.repeat(talos.q0.reshape((-1, 1)), cycle_time * update_rate, axis=1)
    beta = np.linspace(0, 1, traj.shape[1])
    traj[[2, 9, 10, 11, 22, 15, 16, 17, 30]] = (
        0.39 + 0.685 * np.cos(beta),
        -beta,
        2.0 * beta,
        -beta,
        0.1 + beta * 1.56,
        -beta,
        2.0 * beta,
        -beta,
        -0.1 - beta * 1.56,
    )

    if record_video:
        dt = 1 / 60
        fps = 1.0 / dt
        filename = "talos.mp4"
        ctx = talos.viz.create_video_ctx(filename, fps=fps)
        print(f"[video will be recorded @ {filename}]")
    else:
        from contextlib import nullcontext

        ctx = nullcontext()
        print("[no recording]")

    while True:
        with ctx:
            talos.play(traj.T, 1.0 / update_rate)
            traj = np.flip(traj, 1)
            if record_video:
                break


try:
    play_sample_trajectory()
except ViewerClosedError:
    # an exception will be thrown when the window is closed
    pass



import unittest
from math import sqrt

import numpy as np
import pinocchio as pin
from pinocchio.utils import isapprox
from test_case import PinocchioTestCase as TestCase


class TestUtils(TestCase):
    def test_se3ToXYZQUAT_XYZQUATToSe3(self):
        m = pin.SE3.Identity()
        m.translation = np.array([1.0, 2.0, 3.0])
        m.rotation = np.array(
            [[1.0, 0.0, 0.0], [0.0, 0.0, -1.0], [0.0, 1.0, 0.0]]
        )  # rotate('x', pi / 2)
        self.assertApprox(
            pin.SE3ToXYZQUAT(m).T, [1.0, 2.0, 3.0, sqrt(2) / 2, 0, 0, sqrt(2) / 2]
        )
        self.assertApprox(
            pin.XYZQUATToSE3([1.0, 2.0, 3.0, sqrt(2) / 2, 0, 0, sqrt(2) / 2]), m
        )
        self.assertApprox(
            pin.XYZQUATToSE3(np.array([1.0, 2.0, 3.0, sqrt(2) / 2, 0, 0, sqrt(2) / 2])),
            m,
        )
        with self.assertRaises(ValueError):
            pin.XYZQUATToSE3([1.0, 2.0, 3.0])
        with self.assertRaises(ValueError):
            pin.XYZQUATToSE3(np.array([1.0, 2.0, 3.0, sqrt(2) / 2]))

    def test_isapprox(self):
        self.assertFalse(isapprox(1, 2))
        self.assertTrue(isapprox(1, 2, 10))
        self.assertFalse(isapprox([1e10, 1e-7], [1.00001e10, 1e-8]))
        self.assertTrue(isapprox([1e10, 1e-8], [1.00001e10, 1e-9], 1e-5))


if __name__ == "__main__":
    unittest.main()



# Typical header of a Python script using Pinocchio
import gepetto.corbaserver
import pinocchio as pin


# Example of a class Display that connect to Gepetto-viewer and implement a
# 'place' method to set the position/rotation of a 3D visual object in a scene.
class Display:
    """
    Class Display: Example of a class implementing a client for the Gepetto-viewer
    server. The main method of the class is 'place', that sets the position/rotation of
    a 3D visual object in a scene.
    """

    def __init__(self, windowName="pinocchio"):
        """
        This function connect with the Gepetto-viewer server and open a window with the
        given name.  If the window already exists, it is kept in the current state.
        Otherwise, the newly-created window is set up with a scene named 'world'.
        """

        # Create the client and connect it with the display server.
        try:
            self.viewer = gepetto.corbaserver.Client()
        except:  # noqa: E722
            print("Error while starting the viewer client. ")
            print("Check whether Gepetto-viewer is properly started")

        # Open a window for displaying your model.
        try:
            # If the window already exists, do not do anything.
            windowID = self.viewer.gui.getWindowID(windowName)
            print("Warning: window '" + windowName + "' already created.")
            print(
                "The previously created objects will not be destroyed "
                "and do not have to be created again."
            )
        except:  # noqa: E722
            # Otherwise, create the empty window.
            windowID = self.viewer.gui.createWindow(windowName)
            # Start a new "scene" in this window, named "world", with just a floor.
            self.viewer.gui.createScene("world")
            self.viewer.gui.addSceneToWindow("world", windowID)

        # Finally, refresh the layout to obtain your first rendering.
        self.viewer.gui.refresh()

    def place(self, objName, M, refresh=True):
        """
        This function places (ie changes both translation and rotation) of the object
        names "objName" in place given by the SE3 object "M". By default, immediately
        refresh the layout. If multiple objects have to be placed at the same time, do
        the refresh only at the end of the list.
        """
        self.viewer.gui.applyConfiguration(objName, pin.se3ToXYZQUATtuple(M))
        if refresh:
            self.viewer.gui.refresh()



import os
from pathlib import Path

import numpy as np
import pinocchio as pin

np.set_printoptions(linewidth=np.inf)

# ----- PROBLEM STATEMENT ------
#
# We want to find the contact forces and torques required to stand still at a
# configuration 'q0'.
# We assume 3D contacts at each of the feet
#
# The dynamic equation would look like:
#
# M*q_ddot + g(q) + C(q, q_dot) = tau + J^T*lambda --> (for the static case) --> g(q) =
# tau + Jc^T*lambda (1).

# ----- SOLVING STRATEGY ------

# Split the equation between the base link (_bl) joint and the rest of the joints (_j).
# That is,
#
#  | g_bl |   |  0  |   | Jc__feet_bl.T |   | l1 |
#  | g_j  | = | tau | + | Jc__feet_j.T  | * | l2 |    (2)
#                                           | l3 |
#                                           | l4 |

# First, find the contact forces l1, l2, l3, l4 (these are 3 dimensional) by solving for
# the first 6 rows of (2).
# That is,
#
# g_bl   = Jc__feet_bl.T * | l1 |
#                          | l2 |    (3)
#                          | l3 |
#                          | l4 |
#
# Thus we find the contact froces by computing the jacobian pseudoinverse,
#
# | l1 | = pinv(Jc__feet_bl.T) * g_bl  (4)
# | l2 |
# | l3 |
# | l4 |
#
# Now, we can find the necessary torques using the bottom rows in (2). That is,
#
#                             | l1 |
#  tau = g_j - Jc__feet_j.T * | l2 |    (5)
#                             | l3 |
#                             | l4 |

# ----- SOLUTION ------

# 0. DATA
model_path = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
mesh_dir = model_path.parent.parent
urdf_filename = "solo12.urdf"
urdf_model_path = model_path / "solo_description/robots" / urdf_filename

model, collision_model, visual_model = pin.buildModelsFromUrdf(
    urdf_model_path, mesh_dir, pin.JointModelFreeFlyer()
)
data = model.createData()

q0 = np.array(
    [
        0.0,
        0.0,
        0.235,
        0.0,
        0.0,
        0.0,
        1.0,
        0.0,
        0.8,
        -1.6,
        0.0,
        -0.8,
        1.6,
        0.0,
        0.8,
        -1.6,
        0.0,
        -0.8,
        1.6,
    ]
)
v0 = np.zeros(model.nv)
a0 = np.zeros(model.nv)

# 1. GRAVITY TERM

# We compute the gravity terms by using the ID at desired configuration q0, with
# velocity and acceleration being 0. I.e., ID with a = v = 0.
g_grav = pin.rnea(model, data, q0, v0, a0)

g_bl = g_grav[:6]
g_j = g_grav[6:]

# 2. FIND CONTACTS

# First, we set the frame for our contacts. We assume the contacts are placed at the
# following 4 frames and they are 3D.
feet_names = ["FL_FOOT", "FR_FOOT", "HL_FOOT", "HR_FOOT"]
feet_ids = [model.getFrameId(n) for n in feet_names]
bl_id = model.getFrameId("base_link")
ncontact = len(feet_names)

# Now, we need to find the contact Jacobians appearing in (1).
# These are the Jacobians that relate the joint velocity  to the velocity of each feet
Js__feet_q = [
    np.copy(pin.computeFrameJacobian(model, data, q0, id, pin.LOCAL)) for id in feet_ids
]

Js__feet_bl = [np.copy(J[:3, :6]) for J in Js__feet_q]

# Notice that we can write the equation above as an horizontal stack of Jacobians
# transposed and vertical stack of contact forces.
Jc__feet_bl_T = np.zeros([6, 3 * ncontact])
Jc__feet_bl_T[:, :] = np.vstack(Js__feet_bl).T

# Now I only need to do the pinv to compute the contact forces

ls = np.linalg.pinv(Jc__feet_bl_T) @ g_bl  # This is (3)

# Contact forces at local coordinates (at each foot coordinate)
ls__f = np.split(ls, ncontact)

pin.framesForwardKinematics(model, data, q0)

# Contact forces at base link frame
ls__bl = []
for l__f, foot_id in zip(ls__f, feet_ids):
    l_sp__f = pin.Force(l__f, np.zeros(3))
    l_sp__bl = data.oMf[bl_id].actInv(data.oMf[foot_id].act(l_sp__f))
    ls__bl.append(np.copy(l_sp__bl.vector))

print("\n--- CONTACT FORCES ---")
for l__f, foot_id, name in zip(ls__bl, feet_ids, feet_names):
    print(f"Contact force at foot {name} expressed at the BL is: {l__f}")

# Notice that if we add all the contact forces are equal to the g_grav
print(
    "Error between contact forces and gravity at base link: "
    f"{np.linalg.norm(g_bl - sum(ls__bl))}"
)

# 3. FIND TAU
# Find Jc__feet_j
Js_feet_j = [np.copy(J[:3, 6:]) for J in Js__feet_q]

Jc__feet_j_T = np.zeros([12, 3 * ncontact])
Jc__feet_j_T[:, :] = np.vstack(Js_feet_j).T

# Apply (5)
tau = g_j - Jc__feet_j_T @ ls

# 4. CROSS CHECKS

# INVERSE DYNAMICS
# We can compare this torques with the ones one would obtain when computing the ID
# considering the external forces in ls.
pin.framesForwardKinematics(model, data, q0)

joint_names = ["FL_KFE", "FR_KFE", "HL_KFE", "HR_KFE"]
joint_ids = [model.getJointId(n) for n in joint_names]

fs_ext = [pin.Force(np.zeros(6)) for _ in range(len(model.joints))]
for idx, joint in enumerate(model.joints):
    if joint.id in joint_ids:
        fext__bl = pin.Force(ls__bl[joint_ids.index(joint.id)])
        fs_ext[idx] = data.oMi[joint.id].actInv(data.oMf[bl_id].act(fext__bl))

tau_rnea = pin.rnea(model, data, q0, v0, a0, fs_ext)

print("\n--- ID: JOINT TORQUES ---")
print(f"Tau from RNEA:         {tau_rnea}")
print(f"Tau computed manually: {np.append(np.zeros(6), tau)}")
print(f"Tau error: {np.linalg.norm(np.append(np.zeros(6), tau) - tau_rnea)}")

# FORWARD DYNAMICS
# We can also check the results using FD. FD with the tau we got, q0 and v0, should give
# 0 acceleration and the contact forces.
Js_feet3d_q = [np.copy(J[:3, :]) for J in Js__feet_q]
acc = pin.forwardDynamics(
    model,
    data,
    q0,
    v0,
    np.append(np.zeros(6), tau),
    np.vstack(Js_feet3d_q),
    np.zeros(12),
)

print("\n--- FD: ACC. & CONTACT FORCES ---")
print(f"Norm of the FD acceleration: {np.linalg.norm(acc)}")
print(f"Contact forces manually: {ls}")
print(f"Contact forces FD: {data.lambda_c}")
print(f"Contact forces error: {np.linalg.norm(data.lambda_c - ls)}")



import math
import unittest

import numpy as np
import pinocchio as pin
from pinocchio.explog import exp, log
from pinocchio.utils import eye, rand, zero
from test_case import PinocchioTestCase as TestCase


class TestExpLog(TestCase):
    def test_exp3(self):
        v = zero(3)
        m = pin.exp3(v)
        self.assertApprox(m, eye(3))

    def test_Jexp3(self):
        v = zero(3)
        m = pin.Jexp3(v)
        self.assertApprox(m, eye(3))

    def test_log3(self):
        m = eye(3)
        v = pin.log3(m)
        self.assertApprox(v, zero(3))

    def test_log3_quat(self):
        """Test log3 over the quaternions."""
        from eigenpy import Quaternion

        quat_v = np.array([0.0, 0.0, 0.0, 1.0])
        quat = Quaternion(quat_v)
        v0 = pin.log3(quat)
        self.assertApprox(v0, zero(3))

        v1 = pin.log3(quat_v)
        self.assertApprox(v1, zero(3))

    def test_exp3_quat(self):
        self.assertApprox(pin.exp3_quat(zero(3)), np.array([0.0, 0.0, 0.0, 1.0]))

    def test_Jlog3(self):
        m = eye(3)
        J = pin.Jlog3(m)
        self.assertApprox(J, eye(3))

    def test_exp6(self):
        v = pin.Motion.Zero()
        M = pin.exp6(v)
        self.assertTrue(M.isIdentity())

        M2 = pin.exp6(np.array(v))
        self.assertTrue(M2.isIdentity())

    def test_Jexp6(self):
        v = pin.Motion.Zero()
        J = pin.Jexp6(v)
        self.assertApprox(J, eye(6))

        J2 = pin.Jexp6(np.array(v))
        self.assertApprox(J, J2)

    def test_log6(self):
        m = pin.SE3.Identity()
        v = pin.log6(m)
        self.assertApprox(v.vector, zero(6))

    def test_log6_homogeneous(self):
        m = eye(4)
        v = pin.log6(m)
        self.assertApprox(v.vector, zero(6))

    def test_log6_quat(self):
        q0 = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0])
        v = pin.log6_quat(q0)
        self.assertApprox(v.vector, zero(6))

    def test_Jlog6(self):
        m = pin.SE3.Identity()
        J = pin.Jlog6(m)
        self.assertApprox(J, eye(6))

    def test_skew(self):
        u = np.random.rand(3)
        v = np.random.rand(3)

        u_skew = pin.skew(u)
        u_unskew = pin.unSkew(u_skew)

        self.assertApprox(u, u_unskew)

        v_skew = pin.skew(v)
        u_v_square = pin.skewSquare(u, v)

        self.assertApprox(u_v_square, u_skew.dot(v_skew))

    def test_explog(self):
        self.assertApprox(exp(42), math.exp(42))
        self.assertApprox(log(42), math.log(42))
        self.assertApprox(exp(log(42)), 42)
        self.assertApprox(log(exp(42)), 42)

        m = rand(3)
        self.assertLess(np.linalg.norm(m), np.pi)  # necessary for next test
        self.assertApprox(log(exp(m)), m)

        m = np.random.rand(3)
        self.assertLess(np.linalg.norm(m), np.pi)  # necessary for next test
        self.assertApprox(log(exp(m)), m)

        m = pin.SE3.Random()
        self.assertApprox(exp(log(m)), m)

        m = rand(6)
        self.assertLess(
            np.linalg.norm(m), np.pi
        )  # necessary for next test (actually, only angular part)
        self.assertApprox(log(exp(m)), m)

        m = np.random.rand(6)
        self.assertLess(
            np.linalg.norm(m), np.pi
        )  # necessary for next test (actually, only angular part)
        self.assertApprox(log(exp(m)), m)

        m = eye(4)
        self.assertApprox(exp(log(m)).homogeneous, m)

        with self.assertRaises(ValueError):
            exp(eye(4))
        with self.assertRaises(ValueError):
            exp(list(range(3)))
        with self.assertRaises(ValueError):
            log(list(range(3)))
        with self.assertRaises(ValueError):
            log(zero(5))
        with self.assertRaises(ValueError):
            log(zero((3, 1)))


if __name__ == "__main__":
    unittest.main()



import sys
import unittest
from pathlib import Path

sys.path.append(str(Path(__file__).parent.parent))

import casadi
import numpy as np
import pinocchio as pin
import pinocchio.casadi as cpin
from casadi import SX
from test_case import PinocchioTestCase as TestCase


class TestLogExpDerivatives(TestCase):
    def setUp(self) -> None:
        self.cR0 = SX.sym("R0", 3, 3)
        self.cR1 = SX.sym("R1", 3, 3)
        self.dv0 = SX.sym("v0", 3)
        self.dv1 = SX.sym("v1", 3)
        self.v_all = casadi.vertcat(self.dv0, self.dv1)
        self.cv2 = SX.sym("v2", 3)
        self.cdv2 = SX.sym("dv", 3)  # for forming the difference

        self.cR0_i = self.cR0 @ cpin.exp3(self.dv0)
        self.cR1_i = self.cR1 @ cpin.exp3(self.dv1)

        # SE(3) examples
        self.cM0 = SX.sym("M0", 4, 4)
        self.cM1 = SX.sym("M1", 4, 4)
        self.cdw0 = SX.sym("dm0", 6)
        self.cdw1 = SX.sym("dm1", 6)
        self.cw2 = SX.sym("w2", 6)
        self.cdw2 = SX.sym("dw2", 6)
        self.cM0_i = cpin.SE3(self.cM0) * cpin.exp6(self.cdw0)
        self.cM1_i = cpin.SE3(self.cM1) * cpin.exp6(self.cdw1)

    def test_exp3(self):
        """Test the exp map and its derivative."""
        dw = self.cdv2
        exp_expr = self.cR0_i @ cpin.exp3(self.cv2 + dw)

        def repl_dargs(e):
            return casadi.substitute(e, casadi.vertcat(self.dv0, dw), np.zeros(6))

        exp_eval = casadi.Function("exp", [self.cR0, self.cv2], [repl_dargs(exp_expr)])

        diff_expr = cpin.log3(exp_eval(self.cR0, self.cv2).T @ exp_expr)
        Jexp_expr = casadi.jacobian(diff_expr, casadi.vertcat(self.dv0, dw))
        Jexp_eval = casadi.Function(
            "exp", [self.cR0, self.cv2], [repl_dargs(Jexp_expr)]
        )

        w0 = np.zeros(3)
        w1 = np.random.randn(3)
        w2 = np.array([np.pi, 0, 0])
        w3 = np.array([-np.pi, 0, 0])
        R0 = pin.exp3(w0)  # eye(3)
        R1 = pin.exp3(w1)
        R2 = pin.exp3(w2)
        R3 = pin.exp3(w3)
        self.assertApprox(exp_eval(R0, np.zeros(3)).full(), R0)
        self.assertApprox(exp_eval(R0, w1).full(), R1)
        self.assertApprox(exp_eval(R0, -w1).full(), R1.T)
        self.assertApprox(exp_eval(R1, -w1).full(), R0)
        self.assertApprox(exp_eval(R0, w2).full(), R2)

        J0 = pin.Jexp3(w0)
        J1 = pin.Jexp3(w1)
        J2 = pin.Jexp3(w2)
        J3 = pin.Jexp3(w3)
        # print(J0)
        # print(Jexp_eval(R0, np.zeros(3)))
        # print(J1)
        # print(Jexp_eval(R0, w1))
        # print("R1:", R1)
        # print(Jexp_eval(R1, w1))
        # print(Jexp_eval(R2, w1))
        self.assertApprox(Jexp_eval(R0, w0).full(), np.hstack([R0.T, J0]))
        self.assertApprox(Jexp_eval(R0, w1).full(), np.hstack([R1.T, J1]))
        self.assertApprox(Jexp_eval(R0, w2).full(), np.hstack([R2.T, J2]))
        self.assertApprox(Jexp_eval(R0, w3).full(), np.hstack([R3.T, J3]))
        self.assertApprox(Jexp_eval(R1, w0).full(), np.hstack([R0.T, J0]))
        self.assertApprox(Jexp_eval(R1, w2).full(), np.hstack([R2.T, J2]))

    def test_log3(self):
        log_expr = cpin.log3(self.cR0_i.T @ self.cR1_i)
        log_eval = casadi.Function(
            "log",
            [self.cR0, self.cR1],
            [casadi.substitute(log_expr, self.v_all, np.zeros(6))],
        )

        Jlog_expr = casadi.jacobian(log_expr, self.v_all)
        Jlog_eval = casadi.Function(
            "Jlog",
            [self.cR0, self.cR1],
            [casadi.substitute(Jlog_expr, self.v_all, np.zeros(6))],
        )

        R0 = np.eye(3)
        vr = np.random.randn(3)
        R1 = pin.exp3(vr)
        R2 = pin.exp3(np.array([np.pi, 0, 0]))
        v3 = np.array([0, np.pi, 0])
        R3 = pin.exp3(v3)

        self.assertApprox(log_eval(R0, R0).full().squeeze(), np.zeros(3))
        self.assertApprox(log_eval(R0, R1).full().squeeze(), vr)
        self.assertApprox(log_eval(R1, R1).full().squeeze(), np.zeros(3))
        self.assertApprox(log_eval(R0, R2).full().squeeze(), np.array([np.pi, 0, 0]))
        self.assertApprox(log_eval(R0, R3).full().squeeze(), v3)

        J0 = pin.Jlog3(R0)
        jac_identity = np.hstack([-J0, J0])
        self.assertApprox(Jlog_eval(R0, R0).full(), jac_identity)

        J1 = pin.Jlog3(R1)
        self.assertApprox(Jlog_eval(R0, R1).full(), np.hstack([-R1.T @ J1, J1]))

        J2 = pin.Jlog3(R2)
        self.assertApprox(Jlog_eval(R0, R2).full(), np.hstack([-R2.T @ J2, J2]))

    def test_log3_quat(self):
        cquat = SX.sym("quat", 4)
        cdv = SX.sym("dv", 3)
        SO3 = cpin.liegroups.SO3()

        def repl_dargs(e):
            return casadi.substitute(e, cdv, np.zeros(3))

        cquat_i = SO3.integrate(cquat, cdv)

        clog = cpin.log3(cquat_i)
        cJlog = casadi.jacobian(clog, cdv)
        clog_eval = casadi.Function("log", [cquat], [repl_dargs(clog)])
        cJlog_eval = casadi.Function("Jlog", [cquat], [repl_dargs(cJlog)])

        q0 = np.array([0.0, 0.0, 0.0, 1.0])
        q1 = np.array([0.0, 1.0, 0.0, 0.0])
        q2 = np.array([0.0, 0.0, 1.0, 0.0])
        q3 = np.array([1.0, 0.0, 0.0, 0.0])

        self.assertApprox(clog_eval(q0).full().squeeze(), np.zeros(3))
        self.assertApprox(cJlog_eval(q0).full().squeeze(), np.eye(3))

        clog_fun = casadi.dot(clog, clog)
        cJlog_fun = casadi.jacobian(clog_fun, cdv)
        clog_fun_eval = casadi.Function("normlog", [cquat], [repl_dargs(clog_fun)])
        cJlog_fun_eval = casadi.Function("Jnormlog", [cquat], [repl_dargs(cJlog_fun)])

        self.assertApprox(clog_fun_eval(q0).full().squeeze(), np.zeros(1))
        self.assertApprox(cJlog_fun_eval(q0).full(), np.zeros(3))

        self.assertApprox(
            cJlog_fun_eval(q1).full(), 2 * np.pi * np.array([0.0, 1.0, 0.0])
        )
        self.assertApprox(
            cJlog_fun_eval(q2).full(), 2 * np.pi * np.array([0.0, 0.0, 1.0])
        )
        self.assertApprox(
            cJlog_fun_eval(q3).full(), 2 * np.pi * np.array([1.0, 0.0, 0.0])
        )
        print("log3 quat done")

    def test_exp6(self):
        exp_expr = cpin.exp6(self.cw2 + self.cdw2)

        def repl_dargs(e):
            return casadi.substitute(e, self.cdw2, np.zeros(6))

        exp_eval = casadi.Function("exp6", [self.cw2], [repl_dargs(exp_expr.np)])

        diff_expr = cpin.log6(cpin.SE3(exp_eval(self.cw2)).actInv(exp_expr)).np
        Jexp_expr = casadi.jacobian(diff_expr, self.cdw2)

        Jexp_eval = casadi.Function("exp6", [self.cw2], [repl_dargs(Jexp_expr)])

        w0 = np.zeros(6)
        w1 = np.array([0.0, 0.0, 0.0, np.pi, 0.0, 0.0])
        w2 = np.random.randn(6)
        w3 = np.array([0.0, 0.0, 0.0, np.pi / 2, 0.0, 0.0])
        M0 = pin.exp6(w0)
        M1 = pin.exp6(w1)
        M2 = pin.exp6(w2)
        M3 = pin.exp6(w3)
        self.assertApprox(exp_eval(w0).full(), M0.np)
        self.assertApprox(exp_eval(w1).full(), M1.np)
        self.assertApprox(exp_eval(w2).full(), M2.np)
        self.assertApprox(exp_eval(w3).full(), M3.np)

        np.set_printoptions(precision=3)
        J0 = pin.Jexp6(w0)
        self.assertApprox(Jexp_eval(w0).full(), J0)
        J1 = pin.Jexp6(w1)
        self.assertApprox(Jexp_eval(w1).full(), J1)
        J2 = pin.Jexp6(w2)
        self.assertApprox(Jexp_eval(w2).full(), J2)

    def test_log6(self):
        log_expr = cpin.log6(self.cM0_i.actInv(self.cM1_i))

        def repl_dargs(e):
            return casadi.substitute(
                e, casadi.vertcat(self.cdw0, self.cdw1), np.zeros(12)
            )

        log_eval = casadi.Function(
            "log6", [self.cM0, self.cM1], [repl_dargs(log_expr.np)]
        )

        Jlog_expr = casadi.jacobian(log_expr.np, casadi.vertcat(self.cdw0, self.cdw1))
        Jlog_eval = casadi.Function(
            "Jlog6", [self.cM0, self.cM1], [repl_dargs(Jlog_expr)]
        )

        w0 = np.zeros(6)
        M0 = pin.exp6(np.zeros(6))
        w1 = np.array([0, 0, 0, np.pi, 0.0, 0.0])
        M1 = pin.exp6(w1)
        w2 = np.random.randn(6)
        M2 = pin.exp6(w2)

        self.assertApprox(log_eval(M0.np, M0.np).full(), w0)
        self.assertApprox(log_eval(M1.np, M1.np).full(), w0)
        self.assertApprox(log_eval(M0.np, M1.np).full(), w1)
        self.assertApprox(log_eval(M0.np, M2.np).full(), w2)

        J0 = pin.Jlog6(M0)
        J1 = pin.Jlog6(M1)
        J2 = pin.Jlog6(M2)
        self.assertApprox(Jlog_eval(M0.np, M0.np).full(), np.hstack([-J0, J0]))
        self.assertApprox(
            Jlog_eval(M0.np, M1.np).full(), np.hstack([-M1.dualAction.T @ J1, J1])
        )
        self.assertApprox(
            Jlog_eval(M0.np, M2.np).full(), np.hstack([-M2.dualAction.T @ J2, J2])
        )

    def test_log6_quat(self):
        cq0 = SX.sym("q0", 7)
        cv0 = SX.sym("q0", 6)

        def repl_dargs(e):
            return casadi.substitute(e, cv0, np.zeros(6))

        SE3 = cpin.liegroups.SE3()

        cq0_i = SE3.integrate(cq0, cv0)
        clog = cpin.log6_quat(cq0_i).vector
        clog_eval = casadi.Function("log", [cq0], [repl_dargs(clog)])

        cJlog = casadi.jacobian(clog, cv0)
        cJlog_eval = casadi.Function("Jlog", [cq0], [repl_dargs(cJlog)])

        q0 = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0])
        q1 = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0])
        q2 = np.array([0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0])
        q3 = np.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0])
        self.assertApprox(clog_eval(q0).full().squeeze(), np.zeros(6))
        self.assertApprox(cJlog_eval(q0).full(), np.eye(6))

        print("log6 with quats")
        print(clog_eval(q1).full())
        print(cJlog_eval(q1).full())

        clog_fun = casadi.dot(clog, clog)
        clog_fun = casadi.dot(clog, clog)
        cJlog_fun = casadi.jacobian(clog_fun, cv0)
        clog_fun_eval = casadi.Function("normlog", [cq0], [repl_dargs(clog_fun)])
        cJlog_fun_eval = casadi.Function("Jnormlog", [cq0], [repl_dargs(cJlog_fun)])

        print(clog_fun_eval(q0).full().squeeze())
        print(cJlog_fun_eval(q0).full())

        print(clog_fun_eval(q1).full().squeeze())
        print(cJlog_fun_eval(q1).full())

        print(clog_fun_eval(q2).full().squeeze())
        print(cJlog_fun_eval(q2).full())

        print(clog_fun_eval(q3).full().squeeze())
        print(cJlog_fun_eval(q3).full())


if __name__ == "__main__":
    unittest.main()



import gc
import sys

import cpp2pybind11
import pinocchio

a = pinocchio.SE3.Random()
b = pinocchio.SE3.Random()
assert cpp2pybind11.multiply_se3_1(a, b) == a * b
assert cpp2pybind11.multiply_se3(a, b) == a * b
assert cpp2pybind11.no_wrapper.multiply_se3(a, b) == a * b
assert cpp2pybind11.multiply_se3(a) == a
# assert cpp2pybind11.no_wrapper.multiply_se3(a) == a


def print_ref_count(v, what=""):
    # - 2 because one for variable v and one for variable inside getrefcount
    idv = id(v)
    gc.collect()
    # n = len(gc.get_referrers(v))
    n = sys.getrefcount(v)
    print("ref count of", what, idv, n)


m = cpp2pybind11.make_model()
print_ref_count(m, "m")
print(cpp2pybind11.get_ptr(m))
print_ref_count(m, "m")

m.name = ""
cpp2pybind11.testModel1(m)
print_ref_count(m, "m")
assert m.name.startswith("testModel1")

addr2 = cpp2pybind11.testModel2(m, 1)
print_ref_count(m, "m")
assert m.name.startswith("testModel2")

addr3 = cpp2pybind11.testModel3(m, 2)
print_ref_count(m, "m")
assert addr2 == addr3

mm = cpp2pybind11.return_same_model_broken(m)
assert cpp2pybind11.get_ptr(m) != cpp2pybind11.get_ptr(mm)

mm = cpp2pybind11.return_same_model(m)
# Not sure why but the ref count of m and mm sticks to one
print_ref_count(m, "m")
mmm = m
print_ref_count(m, "m")
print_ref_count(mm, "mm")
assert cpp2pybind11.get_ptr(m) == cpp2pybind11.get_ptr(mm)

if False:
    print("deleting m")
    del m
    print("deleted m")
    print("mm is", mm)
else:
    print("deleting mm")
    del mm
    print("deleted mm")
    print("m is", m)



import os
from pathlib import Path
from sys import argv

import pinocchio

model_path = Path(
    Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR")) if len(argv) < 2 else argv[1]
)
mesh_dir = model_path.parent.parent
urdf_model_path = model_path / "ur_description/urdf/ur5_robot.urdf"

# Load the urdf model
model, collision_model, visual_model = pinocchio.buildModelsFromUrdf(
    urdf_model_path, mesh_dir
)
print("model name:", model.name)

# Create data required by the algorithms
data, collision_data, visual_data = pinocchio.createDatas(
    model, collision_model, visual_model
)

# Sample a random configuration
q = pinocchio.randomConfiguration(model)
print(f"q: {q.T}")

# Perform the forward kinematics over the kinematic tree
pinocchio.forwardKinematics(model, data, q)

# Update Geometry models
pinocchio.updateGeometryPlacements(model, data, collision_model, collision_data)
pinocchio.updateGeometryPlacements(model, data, visual_model, visual_data)

# Print out the placement of each joint of the kinematic tree
print("\nJoint placements:")
for name, oMi in zip(model.names, data.oMi):
    print("{:<24} : {: .2f} {: .2f} {: .2f}".format(name, *oMi.translation.T.flat))

# Print out the placement of each collision geometry object
print("\nCollision object placements:")
for k, oMg in enumerate(collision_data.oMg):
    print("{:d} : {: .2f} {: .2f} {: .2f}".format(k, *oMg.translation.T.flat))

# Print out the placement of each visual geometry object
print("\nVisual object placements:")
for k, oMg in enumerate(visual_data.oMg):
    print("{:d} : {: .2f} {: .2f} {: .2f}".format(k, *oMg.translation.T.flat))



import unittest

import pinocchio as pin


class TestFCLTransformConversion(unittest.TestCase):
    def test_from_SE3(self):
        M = pin.SE3.Random()
        fcl_transform = pin.hppfcl.Transform3f(M)

        self.assertTrue((M.rotation == fcl_transform.getRotation()).all())
        self.assertTrue((M.translation == fcl_transform.getTranslation()).all())

    def test_to_SE3(self):
        fcl_transform = pin.hppfcl.Transform3f()
        M = pin.SE3(fcl_transform)
        self.assertTrue(M.isIdentity())


if __name__ == "__main__":
    if pin.WITH_HPP_FCL_BINDINGS:
        unittest.main()



from math import pi

import numpy as np
import pinocchio as pin
from display import Display
from numpy.linalg import norm, pinv
from pinocchio.utils import cross, eye, rotate, zero


class Visual:
    """
    Class representing one 3D mesh of the robot, to be attached to a joint. The class
    contains:
    * the name of the 3D objects inside Gepetto viewer.
    * the ID of the joint in the kinematic tree to which the body is attached.
    * the placement of the body with respect to the joint frame.
    This class is only used in the list Robot.visuals (see below).

    The visual are supposed mostly to be capsules. In that case, the object also
    contains radius and length of the capsule.
    The collision checking computes collision test, distance, and witness points.
    Using the supporting robot, the collision Jacobian returns a 1xN matrix
    corresponding to the normal direction.
    """

    def __init__(self, name, jointParent, placement, radius=0.1, length=None):
        """Length and radius are used in case of capsule objects"""
        self.name = name  # Name in gepetto viewer
        self.jointParent = jointParent  # ID (int) of the joint
        self.placement = placement  # placement of the body wrt joint, i.e. bodyMjoint
        if length is not None:
            self.length = length
            self.radius = radius

    def place(self, display, oMjoint):
        oMbody = oMjoint * self.placement
        display.place(self.name, oMbody, False)

    def isCapsule(self):
        return hasattr(self, "length") and hasattr(self, "radius")

    def collision(self, c2, data=None, oMj1=None, oMj2=None):
        if data is not None:
            oMj1 = data.oMi[self.jointParent]
            oMj2 = data.oMi[c2.jointParent]
        M1 = oMj1 * self.placement
        M2 = oMj2 * c2.placement
        assert self.isCapsule() and c2.isCapsule()
        l1 = self.length
        r1 = self.radius
        l2 = c2.length
        r2 = c2.radius

        a1 = M1.act(np.array([0, 0, -l1 / 2]))
        b1 = M2.act(np.array([0, 0, -l2 / 2]))
        a2 = M1.act(np.array([0, 0, +l1 / 2]))
        b2 = M2.act(np.array([0, 0, +l2 / 2]))

        a1a2_b1b2 = np.array([a1 - a2, b1 - b2]).T
        ab = pinv(a1a2_b1b2) * (b2 - a2)

        if (0 <= ab).all() and (ab <= 1).all():
            asat = bsat = False
            pa = a2 + ab[0, 0] * (a1 - a2)
            pb = b2 + ab[1, 0] * (b1 - b2)
        else:
            asat = bsat = True
            i = np.argmin(np.vstack([ab, 1 - ab]))

            pa = a2 if i == 0 else a1
            pb = b2 if i == 1 else b1
            if i == 0 or i == 2:  # fix a to pa, search b
                b = (pinv((b1 - b2).reshape(3, 1)) * (pa - b2))[0, 0]
                if b < 0:
                    pb = b2
                elif b > 1:
                    pb = b1
                else:
                    pb = b2 + b * (b1 - b2)
                    bsat = False
            else:  # fix b
                a = (pinv((a1 - a2).reshape(3, 1)) * (pb - a2))[0, 0]
                if a < 0:
                    pa = a2
                elif a > 1:
                    pa = a1
                else:
                    pa = a2 + a * (a1 - a2)
                    asat = False

        print(asat, bsat)
        dist = norm(pa - pb) - (r1 + r2)
        if norm(pa - pb) > 1e-3:
            # Compute witness points
            ab = pa - pb
            ab /= norm(ab)
            wa = pa - ab * r1
            wb = pb + ab * r2

            # Compute normal matrix
            x = np.array([1.0, 0, 0])
            r1 = cross(ab, x)
            if norm(r1) < 1e-2:
                x = np.array([0, 1.0, 0])
            r1 = cross(ab, x)
            r1 /= norm(r1)
            r2 = cross(ab, r1)
            R = np.hstack([r1, r2, ab.reshape(3, 1)])

            self.dist = dist
            c2.dist = dist
            self.w = wa
            c2.w = wb
            self.R = R
            c2.R = R

        return dist

    def jacobian(self, c2, robot, q):
        Ja = pin.jacobian(robot.model, robot.data, q, self.jointParent, False, True)
        Jb = pin.jacobian(robot.model, robot.data, q, c2.jointParent, False, True)

        Xa = pin.SE3(self.R, self.w).action
        Xb = pin.SE3(c2.R, c2.w).action

        J = (Xa * Ja)[2, :] - (Xb * Jb)[2, :]
        return J

    def displayCollision(self, viewer, name="world/wa"):
        viewer.viewer.gui.setVisibility(name, "ON")
        viewer.place(name, pin.SE3(self.R, self.w))


class Robot:
    """
    Define a class Robot with 7DOF (shoulder=3 + elbow=1 + wrist=3).
    The configuration is nq=7. The velocity is the same.
    The members of the class are:
    * viewer: a display encapsulating a gepetto viewer client to create 3D objects and
      place them.
    * model: the kinematic tree of the robot.
    * data: the temporary variables to be used by the kinematic algorithms.
    * visuals: the list of all the 'visual' 3D objects to render the robot, each element
      of the list being an object Visual (see above).

    CollisionPairs is a list of visual indexes.
    Reference to the collision pair is used in the collision test and jacobian of the
    collision (which are simply proxy method to methods of the visual class).
    """

    def __init__(self):
        self.viewer = Display()
        self.visuals = []
        self.model = pin.Model()
        self.createHand()
        self.data = self.model.createData()
        self.q0 = zero(self.model.nq)
        # self.q0[3] = 1.0
        self.v0 = zero(self.model.nv)
        self.collisionPairs = []

    def createHand(self, root_id=0, prefix="", joint_placement=None):
        def trans(x, y, z):
            return pin.SE3(eye(3), np.array([x, y, z]))

        def inertia(m, c):
            return pin.Inertia(m, np.array(c, np.double), eye(3) * m**2)

        def joint_name(body):
            return prefix + body + "_joint"

        def body_name(body):
            return "world/" + prefix + body

        color = [red, green, blue, transparency] = [1, 1, 0.78, 1.0]
        joint_id = root_id
        cm = 1e-2

        joint_placement = (
            joint_placement if joint_placement is not None else pin.SE3.Identity()
        )
        joint_id = self.model.addJoint(
            joint_id, pin.JointModelRY(), joint_placement, joint_name("wrist")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(3, [0, 0, 0]), pin.SE3.Identity()
        )

        L, W, H = 3 * cm, 5 * cm, 1 * cm
        self.viewer.viewer.gui.addSphere(body_name("wrist"), 0.02, color)
        self.viewer.viewer.gui.addBox(body_name("wpalm"), L / 2, W / 2, H, color)
        self.visuals.append(Visual(body_name("wpalm"), joint_id, trans(L / 2, 0, 0)))

        self.viewer.viewer.gui.addCapsule(body_name("wpalmb"), H, W, color)
        self.visuals.append(
            Visual(
                body_name("wpalmb"),
                joint_id,
                pin.SE3(rotate("x", pi / 2), zero(3)),
                H,
                W,
            )
        )

        self.viewer.viewer.gui.addCapsule(body_name("wpalmt"), H, W, color)
        pos = pin.SE3(rotate("x", pi / 2), np.array([L, 0, 0]))
        self.visuals.append(Visual(body_name("wpalmt"), joint_id, pos, H, W))

        self.viewer.viewer.gui.addCapsule(body_name("wpalml"), H, L, color)
        pos = pin.SE3(rotate("y", pi / 2), np.array([L / 2, -W / 2, 0]))
        self.visuals.append(Visual(body_name("wpalml"), joint_id, pos, H, L))

        self.viewer.viewer.gui.addCapsule(body_name("wpalmr"), H, L, color)
        pos = pin.SE3(rotate("y", pi / 2), np.array([L / 2, +W / 2, 0]))
        self.visuals.append(Visual(body_name("wpalmr"), joint_id, pos, H, L))

        joint_placement = pin.SE3(eye(3), np.array([5 * cm, 0, 0]))
        joint_id = self.model.addJoint(
            joint_id, pin.JointModelRY(), joint_placement, joint_name("palm")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(2, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addCapsule(body_name("palm2"), 1 * cm, W, color)
        self.visuals.append(
            Visual(
                body_name("palm2"),
                joint_id,
                pin.SE3(rotate("x", pi / 2), zero(3)),
                H,
                W,
            )
        )

        FL = 4 * cm
        palmIdx = joint_id

        joint_placement = pin.SE3(eye(3), np.array([2 * cm, W / 2, 0]))
        joint_id = self.model.addJoint(
            palmIdx, pin.JointModelRY(), joint_placement, joint_name("finger11")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.5, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addCapsule(body_name("finger11"), H, FL - 2 * H, color)
        pos = pin.SE3(rotate("y", pi / 2), np.array([FL / 2 - H, 0, 0]))
        self.visuals.append(Visual(body_name("finger11"), joint_id, pos, H, FL - 2 * H))

        joint_placement = pin.SE3(eye(3), np.array([FL, 0, 0]))
        joint_id = self.model.addJoint(
            joint_id, pin.JointModelRY(), joint_placement, joint_name("finger12")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.5, [0, 0, 0]), pin.SE3.Identity()
        )

        self.viewer.viewer.gui.addCapsule(body_name("finger12"), H, FL - 2 * H, color)
        pos = pin.SE3(rotate("y", pi / 2), np.array([FL / 2 - H, 0, 0]))
        self.visuals.append(Visual(body_name("finger12"), joint_id, pos, H, FL - 2 * H))

        joint_placement = pin.SE3(eye(3), np.array([FL - 2 * H, 0, 0]))
        joint_id = self.model.addJoint(
            joint_id, pin.JointModelRY(), joint_placement, joint_name("finger13")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.3, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addSphere(body_name("finger13"), H, color)
        self.visuals.append(
            Visual(body_name("finger13"), joint_id, trans(2 * H, 0, 0), H, 0)
        )

        joint_placement = pin.SE3(eye(3), np.array([2 * cm, 0, 0]))
        joint_id = self.model.addJoint(
            palmIdx, pin.JointModelRY(), joint_placement, joint_name("finger21")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.5, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addCapsule(body_name("finger21"), H, FL - 2 * H, color)
        pos = pin.SE3(rotate("y", pi / 2), np.array([FL / 2 - H, 0, 0]))
        self.visuals.append(Visual(body_name("finger21"), joint_id, pos, H, FL - 2 * H))

        joint_placement = pin.SE3(eye(3), np.array([FL, 0, 0]))
        joint_id = self.model.addJoint(
            joint_id, pin.JointModelRY(), joint_placement, joint_name("finger22")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.5, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addCapsule(body_name("finger22"), H, FL - 2 * H, color)
        pos = pin.SE3(rotate("y", pi / 2), np.array([FL / 2 - H, 0, 0]))
        self.visuals.append(Visual(body_name("finger22"), joint_id, pos, H, FL - 2 * H))

        joint_placement = pin.SE3(eye(3), np.array([FL - H, 0, 0]))
        joint_id = self.model.addJoint(
            joint_id, pin.JointModelRY(), joint_placement, joint_name("finger23")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.3, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addSphere(body_name("finger23"), H, color)
        self.visuals.append(
            Visual(body_name("finger23"), joint_id, trans(H, 0, 0), H, 0)
        )

        joint_placement = pin.SE3(eye(3), np.array([2 * cm, -W / 2, 0]))
        joint_id = self.model.addJoint(
            palmIdx, pin.JointModelRY(), joint_placement, joint_name("finger31")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.5, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addCapsule(body_name("finger31"), H, FL - 2 * H, color)
        pos = pin.SE3(rotate("y", pi / 2), np.array([FL / 2 - H, 0, 0]))
        self.visuals.append(Visual(body_name("finger31"), joint_id, pos, H, FL - 2 * H))

        joint_placement = pin.SE3(eye(3), np.array([FL, 0, 0]))
        joint_id = self.model.addJoint(
            joint_id, pin.JointModelRY(), joint_placement, joint_name("finger32")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.5, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addCapsule(body_name("finger32"), H, FL - 2 * H, color)
        pos = pin.SE3(rotate("y", pi / 2), np.array([FL / 2 - H, 0, 0]))
        self.visuals.append(Visual(body_name("finger32"), joint_id, pos, H, FL - 2 * H))

        joint_placement = pin.SE3(eye(3), np.array([FL - 2 * H, 0, 0]))
        joint_id = self.model.addJoint(
            joint_id, pin.JointModelRY(), joint_placement, joint_name("finger33")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.3, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addSphere(body_name("finger33"), H, color)
        self.visuals.append(
            Visual(body_name("finger33"), joint_id, trans(2 * H, 0, 0), H, 0)
        )

        joint_placement = pin.SE3(eye(3), np.array([1 * cm, -W / 2 - H * 1.5, 0]))
        joint_id = self.model.addJoint(
            1, pin.JointModelRY(), joint_placement, joint_name("thumb1")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.5, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addCapsule(body_name("thumb1"), H, 2 * cm, color)
        pos = pin.SE3(
            rotate("z", pi / 3) * rotate("x", pi / 2), np.array([1 * cm, -1 * cm, 0])
        )
        self.visuals.append(Visual(body_name("thumb1"), joint_id, pos, 2 * cm))

        joint_placement = pin.SE3(
            rotate("z", pi / 3) * rotate("x", pi), np.array([3 * cm, -1.8 * cm, 0])
        )
        joint_id = self.model.addJoint(
            joint_id, pin.JointModelRZ(), joint_placement, joint_name("thumb2")
        )
        self.model.appendBodyToJoint(
            joint_id, inertia(0.4, [0, 0, 0]), pin.SE3.Identity()
        )
        self.viewer.viewer.gui.addCapsule(body_name("thumb2"), H, FL - 2 * H, color)
        pos = pin.SE3(rotate("x", pi / 3), np.array([-0.7 * cm, 0.8 * cm, -0.5 * cm]))
        self.visuals.append(Visual(body_name("thumb2"), joint_id, pos, H, FL - 2 * H))

        # Prepare some patches to represent collision points. Yet unvisible.
        for i in range(10):
            self.viewer.viewer.gui.addCylinder(
                f"world/wa{i}", 0.01, 0.003, [1.0, 0, 0, 1]
            )
            self.viewer.viewer.gui.addCylinder(
                f"world/wb{i}", 0.01, 0.003, [1.0, 0, 0, 1]
            )
            self.viewer.viewer.gui.setVisibility(f"world/wa{i}", "OFF")
            self.viewer.viewer.gui.setVisibility(f"world/wb{i}", "OFF")

    def checkCollision(self, pairIndex):
        ia, ib = self.collisionPairs[pairIndex]
        va = self.visuals[ia]
        vb = self.visuals[ib]
        dist = va.collision(vb, self.data)
        return dist

    def collisionJacobian(self, pairIndex, q):
        ia, ib = self.collisionPairs[pairIndex]
        va = self.visuals[ia]
        vb = self.visuals[ib]
        return va.jacobian(vb, self, q)

    def displayCollision(self, pairIndex, meshIndex, onlyOne=False):
        ia, ib = self.collisionPairs[pairIndex]
        va = self.visuals[ia]
        vb = self.visuals[ib]
        va.displayCollision(self.viewer, f"world/wa{meshIndex}")
        vb.displayCollision(self.viewer, f"world/wb{meshIndex}")
        self.viewer.viewer.gui.setVisibility(f"world/wa{meshIndex}", "ON")
        self.viewer.viewer.gui.setVisibility(f"world/wb{meshIndex}", "ON")

    def display(self, q):
        pin.forwardKinematics(self.model, self.data, q)
        for visual in self.visuals:
            visual.place(self.viewer, self.data.oMi[visual.jointParent])
        self.viewer.viewer.gui.refresh()



import numpy as np
import pinocchio as pin
import pinocchio.cppadcg as cgpin
from pycppad import (
    ADCG,
    CG,
    ADCGFun,
    CodeHandler,
    Independent,
    LangCDefaultVariableNameGenerator,
    LanguageC,
)

pinmodel = pin.buildSampleModelHumanoidRandom()
model = cgpin.Model(pinmodel)
data = model.createData()

nq = model.nq
nv = model.nv

x = np.array([ADCG(CG(0.0))] * (nq + nv + nv))
x[:nq] = cgpin.neutral(model)
Independent(x)

y = cgpin.rnea(model, data, x[:nq], x[nq : nq + nv], x[nq + nv :])

fun = ADCGFun(x, y)

# /***************************************************************************
# *                        Generate the C source code
# **************************************************************************/

# /**
# * start the special steps for source code generation for a Jacobian
# */
handler = CodeHandler(50)

indVars = np.array([CG(1.0)] * (nq + nv + nv))
handler.makeVariables(indVars)

jac = fun.Jacobian(indVars)

langC = LanguageC("double", 3)
nameGen = LangCDefaultVariableNameGenerator("y", "x", "v", "array", "sarray")
code = handler.generateCode(langC, jac, nameGen, "source")
print(code)



import warnings

import numpy as np

from .. import pinocchio_pywrap_default as pin
from ..utils import npToTuple
from . import BaseVisualizer

try:
    import hppfcl

    WITH_HPP_FCL_BINDINGS = True
except ImportError:
    WITH_HPP_FCL_BINDINGS = False


def create_capsule_markers(marker_ref, oMg, d, fl):
    """Make capsule using two sphere and one cylinder"""
    from copy import deepcopy

    from geometry_msgs.msg import Point
    from visualization_msgs.msg import Marker

    displacment = pin.SE3.Identity()

    displacment.translation[2] = fl / 2.0
    oMsphere_1 = oMg * displacment
    displacment.translation[2] = -fl / 2.0
    oMsphere_2 = oMg * displacment

    marker_cylinder = marker_ref
    marker_cylinder.type = Marker.CYLINDER
    marker_cylinder.scale = Point(d, d, fl)
    marker_cylinder.pose = SE3ToROSPose(oMg)

    marker_sphere_1 = deepcopy(marker_ref)
    marker_sphere_1.id += 10000  # How to ensure this id is not taken ?
    marker_sphere_1.type = Marker.SPHERE
    marker_sphere_1.scale = Point(d, d, d)
    marker_sphere_1.pose = SE3ToROSPose(oMsphere_1)

    marker_sphere_2 = deepcopy(marker_ref)
    marker_sphere_2.id += 20000  # How to ensure this id is not taken ?
    marker_sphere_2.type = Marker.SPHERE
    marker_sphere_2.scale = Point(d, d, d)
    marker_sphere_2.pose = SE3ToROSPose(oMsphere_2)

    return [marker_cylinder, marker_sphere_1, marker_sphere_2]


def SE3ToROSPose(oMg):
    """Converts SE3 matrix to ROS geometry_msgs/Pose format"""
    from geometry_msgs.msg import Point, Pose, Quaternion

    xyz_quat = pin.SE3ToXYZQUATtuple(oMg)
    return Pose(position=Point(*xyz_quat[:3]), orientation=Quaternion(*xyz_quat[3:]))


class RVizVisualizer(BaseVisualizer):
    """A Pinocchio display using RViz"""

    class Viewer:
        app = None
        viz = None
        viz_manager = None

    def initViewer(
        self,
        viewer=None,
        windowName="python-pinocchio",
        loadModel=False,
        initRosNode=True,
    ):
        """
        Init RVizViewer by starting a ros node (or not) and creating an RViz window.
        """
        from python_qt_binding.QtWidgets import QApplication
        from rosgraph import is_master_online
        from rospy import WARN, init_node
        from rviz import bindings as rviz

        if not is_master_online():  # Checks the master uri
            # ROS Master is offline
            warnings.warn(
                "Error while importing the viz client.\n"
                "Check whether ROS master (roscore) is properly started",
                category=UserWarning,
                stacklevel=2,
            )
            return None

        if initRosNode:
            init_node("pinocchio_viewer", anonymous=True, log_level=WARN)

        if viewer is None:
            self.viewer = RVizVisualizer.Viewer()
            self.viewer.app = QApplication([])
            self.viewer.viz = rviz.VisualizationFrame()
            self.viewer.viz.setSplashPath("")
            self.viewer.viz.initialize()
            self.viewer.viz.setWindowTitle(windowName + "[*]")
            self.viewer.viz_manager = self.viewer.viz.getManager()
            self.viewer.viz.show()
        else:
            self.viewer = viewer

        if loadModel:
            self.loadViewerModel()

        return self.viewer

    def loadViewerModel(self, rootNodeName="pinocchio"):
        """Create the displays in RViz and create publishers for the MarkerArray"""
        from rospy import Publisher
        from visualization_msgs.msg import MarkerArray

        # Visuals
        self.visuals_publisher = Publisher(
            rootNodeName + "_visuals", MarkerArray, queue_size=1, latch=True
        )
        self.visual_Display = self.viewer.viz_manager.createDisplay(
            "rviz/MarkerArray", rootNodeName + "_visuals", True
        )
        self.visual_Display.subProp("Marker Topic").setValue(rootNodeName + "_visuals")
        self.visual_ids = []

        # Collisions
        self.collisions_publisher = Publisher(
            rootNodeName + "_collisions", MarkerArray, queue_size=1, latch=True
        )
        self.collision_Display = self.viewer.viz_manager.createDisplay(
            "rviz/MarkerArray", rootNodeName + "_collisions", True
        )
        self.collision_Display.subProp("Marker Topic").setValue(
            rootNodeName + "_collisions"
        )
        self.collision_ids = []

        # Group
        root_group = self.viewer.viz_manager.getRootDisplayGroup()
        self.group_Display = self.viewer.viz_manager.createDisplay(
            "rviz/Group", rootNodeName, True
        )
        self.group_Display.addChild(
            root_group.takeChild(self.visual_Display)
        )  # Remove display from root group and add it to robot group
        self.group_Display.addChild(
            root_group.takeChild(self.collision_Display)
        )  # Remove display from root group and add it to robot group

        self.seq = 0
        self.display()

    def clean(self):
        """Delete all the objects from the whole scene"""
        if hasattr(self, "collisions_publisher"):
            self._clean(self.collisions_publisher)
            self.collision_ids = []

        if hasattr(self, "visuals_publisher"):
            self._clean(self.visuals_publisher)
            self.visual_ids = []

    def display(self, q=None):
        """Display the robot at configuration q in the viz by placing all the bodies."""
        # Update the robot kinematics and geometry.
        if q is not None:
            pin.forwardKinematics(self.model, self.data, q)

        if self.collision_model is not None and hasattr(self, "collisions_publisher"):
            pin.updateGeometryPlacements(
                self.model, self.data, self.collision_model, self.collision_data
            )
            self.collision_ids = self._plot(
                self.collisions_publisher,
                self.collision_model,
                self.collision_data,
                self.collision_ids,
            )

        if self.visual_model is not None and hasattr(self, "visuals_publisher"):
            pin.updateGeometryPlacements(
                self.model, self.data, self.visual_model, self.visual_data
            )
            self.visual_ids = self._plot(
                self.visuals_publisher,
                self.visual_model,
                self.visual_data,
                self.visual_ids,
            )

    def _plot(self, publisher, model, data, previous_ids=()):
        """
        Create markers for each object of the model and publish it as MarkerArray
        (also delete unused previously created markers)
        """
        from geometry_msgs.msg import Point
        from rospy import get_rostime
        from std_msgs.msg import ColorRGBA, Header
        from visualization_msgs.msg import Marker, MarkerArray

        self.seq += 1
        header = Header(
            frame_id="map", seq=self.seq, stamp=get_rostime()
        )  # Common header for every marker

        marker_array = MarkerArray()
        for obj in model.geometryObjects:
            obj_id = model.getGeometryId(obj.name)

            # Prepare marker
            marker = Marker()
            marker.id = obj_id
            marker.header = header
            marker.action = Marker.ADD  # same as Marker.MODIFY
            marker.pose = SE3ToROSPose(data.oMg[obj_id])
            marker.color = ColorRGBA(*obj.meshColor)

            if obj.meshTexturePath != "":
                warnings.warn(
                    "Textures are not supported in RVizVisualizer (for "
                    + obj.name
                    + ")",
                    category=UserWarning,
                    stacklevel=2,
                )

            # Create geometry
            geom = obj.geometry
            if WITH_HPP_FCL_BINDINGS and isinstance(geom, hppfcl.ShapeBase):
                # append a primitive geometry
                if isinstance(geom, hppfcl.Cylinder):
                    d, fl = 2 * geom.radius, 2 * geom.halfLength
                    marker.type = Marker.CYLINDER
                    marker.scale = Point(d, d, fl)
                    marker_array.markers.append(marker)
                elif isinstance(geom, hppfcl.Box):
                    size = npToTuple(2.0 * geom.halfSide)
                    marker.type = Marker.CUBE
                    marker.scale = Point(*size)
                    marker_array.markers.append(marker)
                elif isinstance(geom, hppfcl.Sphere):
                    d = 2 * geom.radius
                    marker.type = Marker.SPHERE
                    marker.scale = Point(d, d, d)
                    marker_array.markers.append(marker)
                elif isinstance(geom, hppfcl.Capsule):
                    d, fl = 2 * geom.radius, 2 * geom.halfLength
                    marker_array.markers.extend(
                        create_capsule_markers(marker, data.oMg[obj_id], d, fl)
                    )
                else:
                    msg = f"Unsupported geometry type for {obj.name} ({type(geom)})"
                    warnings.warn(msg, category=UserWarning, stacklevel=2)
                    continue
            else:
                # append a mesh
                marker.type = Marker.MESH_RESOURCE  # Custom mesh
                marker.scale = Point(*npToTuple(obj.meshScale))
                marker.mesh_resource = "file://" + obj.meshPath
                marker_array.markers.append(marker)

        # Remove unused markers
        new_ids = [marker.id for marker in marker_array.markers]
        for old_id in previous_ids:
            if old_id not in new_ids:
                marker_remove = Marker()
                marker_remove.header = header
                marker_remove.id = old_id
                marker_remove.action = Marker.DELETE
                marker_array.markers.append(marker_remove)

        # Publish markers
        publisher.publish(marker_array)

        # Return list of markers id
        return new_ids

    def _clean(self, publisher):
        """Delete all the markers from a topic (use one marker with action DELETEALL)"""
        from rospy import get_rostime
        from std_msgs.msg import Header
        from visualization_msgs.msg import Marker, MarkerArray

        # Increment seq number
        self.seq += 1

        # Prepare a clean_all marker
        marker = Marker()
        marker.header = Header(frame_id="map", seq=self.seq, stamp=get_rostime())
        marker.action = Marker.DELETEALL

        # Add the marker to a MarkerArray
        marker_array = MarkerArray(markers=[marker])

        # Publish marker
        publisher.publish(marker_array)

    def displayCollisions(self, visibility):
        """Set whether to display collision objects or not"""
        self.collision_Display.setEnabled(visibility)

    def displayVisuals(self, visibility):
        """Set whether to display visual objects or not"""
        self.visual_Display.setEnabled(visibility)

    def sleep(self, dt):
        from python_qt_binding.QtTest import QTest

        QTest.qWait(1e3 * dt)

    def setBackgroundColor(self):
        raise NotImplementedError()

    def setCameraTarget(self, target):
        raise NotImplementedError()

    def setCameraPosition(self, position: np.ndarray):
        raise NotImplementedError()

    def setCameraZoom(self, zoom: float):
        raise NotImplementedError()

    def setCameraPose(self, pose: np.ndarray):
        raise NotImplementedError()

    def captureImage(self, w=None, h=None):
        raise NotImplementedError()

    def disableCameraControl(self):
        raise NotImplementedError()

    def enableCameraControl(self):
        raise NotImplementedError()

    def drawFrameVelocities(self, *args, **kwargs):
        raise NotImplementedError()


__all__ = ["RVizVisualizer"]



# This examples shows how to load and move a robot in meshcat.
# Note: this feature requires Meshcat to be installed, this can be done using
# pip install --user meshcat

import sys
import time

import hppfcl as fcl
import numpy as np
import pinocchio as pin
from pinocchio.visualize import MeshcatVisualizer

if tuple(map(int, fcl.__version__.split("."))) >= (3, 0, 0):
    with_octomap = fcl.WITH_OCTOMAP
else:
    with_octomap = False
if not with_octomap:
    print(
        "This example is skiped as HPP-FCL has not been compiled with octomap support."
    )

model = pin.Model()
collision_model = pin.GeometryModel()

octree = fcl.makeOctree(np.random.rand(1000, 3), 0.01)
octree_object = pin.GeometryObject("octree", 0, pin.SE3.Identity(), octree)
octree_object.meshColor[0] = 1.0
collision_model.addGeometryObject(octree_object)

visual_model = collision_model
viz = MeshcatVisualizer(model, collision_model, visual_model)

# Start a new MeshCat server and client.
# Note: the server can also be started separately using the "meshcat-server" command in
# a terminal:
# this enables the server to remain active after the current script ends.
#
# Option open=True pens the visualizer.
# Note: the visualizer can also be opened seperately by visiting the provided URL.
try:
    viz.initViewer(open=True)
except ImportError as err:
    print(
        "Error while initializing the viewer. "
        "It seems you should install Python meshcat"
    )
    print(err)
    sys.exit(0)

viz.loadViewerModel()
time.sleep(1.0)



# TODO: Remove when py39 is dropped
# This allow to use new Optional syntax
from __future__ import annotations

import warnings
from pathlib import Path
from typing import ClassVar, Union

import numpy as np

from .. import pinocchio_pywrap_default as pin
from ..deprecation import DeprecatedWarning
from ..utils import npToTuple
from . import BaseVisualizer

try:
    import meshcat
    import meshcat.geometry as mg
except ImportError:
    import_meshcat_succeed = False
else:
    import_meshcat_succeed = True

import base64

# DaeMeshGeometry
import xml.etree.ElementTree as Et
from typing import Any

# TODO: Use Union syntax when py39 is dropped
MsgType = dict[str, Union[str, bytes, bool, float, "MsgType"]]

try:
    import hppfcl

    WITH_HPP_FCL_BINDINGS = True
except ImportError:
    WITH_HPP_FCL_BINDINGS = False

DEFAULT_COLOR_PROFILES = {
    "gray": ([0.98, 0.98, 0.98], [0.8, 0.8, 0.8]),
    "white": ([1.0, 1.0, 1.0], [1.0, 1.0, 1.0]),
}
COLOR_PRESETS = DEFAULT_COLOR_PROFILES.copy()

FRAME_AXIS_POSITIONS = (
    np.array([[0, 0, 0], [1, 0, 0], [0, 0, 0], [0, 1, 0], [0, 0, 0], [0, 0, 1]])
    .astype(np.float32)
    .T
)
FRAME_AXIS_COLORS = (
    np.array([[1, 0, 0], [1, 0.6, 0], [0, 1, 0], [0.6, 1, 0], [0, 0, 1], [0, 0.6, 1]])
    .astype(np.float32)
    .T
)


def getColor(color):
    assert color is not None
    color = np.asarray(color)
    assert color.shape == (3,)
    return color.clip(0.0, 1.0)


def hasMeshFileInfo(geometry_object):
    """Check whether the geometry object contains a Mesh supported by MeshCat"""
    if geometry_object.meshPath == "":
        return False

    file_extension = Path(geometry_object.meshPath).suffix
    if file_extension.lower() in [".dae", ".obj", ".stl"]:
        return True

    return False


def applyScalingOnHomegeneousTransform(
    homogeneous_transform: np.ndarray, scale: np.ndarray
) -> np.ndarray:
    assert homogeneous_transform.shape == (4, 4)
    assert scale.size == 3
    scale = np.array(scale).flatten()
    S = np.diag(np.concatenate((scale, [1.0])))
    return homogeneous_transform @ S


if import_meshcat_succeed:
    # Code adapted from Jiminy
    class Cone(mg.Geometry):
        """A cone of the given height and radius. By Three.js convention, the axis
        of rotational symmetry is aligned with the y-axis.
        """

        def __init__(
            self,
            height: float,
            radius: float,
            radialSegments: float = 32,
            openEnded: bool = False,
        ):
            super().__init__()
            self.radius = radius
            self.height = height
            self.radialSegments = radialSegments
            self.openEnded = openEnded

        def lower(self, object_data: Any) -> MsgType:
            return {
                "uuid": self.uuid,
                "type": "ConeGeometry",
                "radius": self.radius,
                "height": self.height,
                "radialSegments": self.radialSegments,
                "openEnded": self.openEnded,
            }

    class DaeMeshGeometry(mg.ReferenceSceneElement):
        def __init__(self, dae_path: str, cache: set[str] | None = None) -> None:
            """Load Collada files with texture images.
            Inspired from
            https://gist.github.com/danzimmerman/a392f8eadcf1166eb5bd80e3922dbdc5
            """
            # Init base class
            super().__init__()

            dae_path = Path(dae_path)

            # Attributes to be specified by the user
            self.path = None
            self.material = None
            self.intrinsic_transform = mg.tf.identity_matrix()

            # Raw file content
            dae_dir = dae_path.parent
            with dae_path.open() as text_file:
                self.dae_raw = text_file.read()

            # Parse the image resource in Collada file
            img_resource_paths: list[Path] = []
            img_lib_element = Et.parse(dae_path).find(
                "{http://www.collada.org/2005/11/COLLADASchema}library_images"
            )
            if img_lib_element:
                img_resource_paths = [
                    Path(e.text)
                    for e in img_lib_element.iter()
                    if e.tag.count("init_from")
                ]

            # Convert textures to data URL for Three.js ColladaLoader to load them
            self.img_resources: dict[str, str] = {}
            for img_path in img_resource_paths:
                img_key = str(img_path)
                # Return empty string if already in cache
                if cache is not None:
                    if img_path in cache:
                        self.img_resources[img_key] = ""
                        continue
                    cache.add(img_path)

                # Encode texture in base64
                img_path_abs: Path = img_path
                if not img_path.is_absolute():
                    img_path_abs = (dae_dir / img_path_abs).resolve()
                if not img_path_abs.is_file():
                    raise UserWarning(f"Texture '{img_path}' not found.")
                with img_path_abs.open("rb") as img_file:
                    img_data = base64.b64encode(img_file.read())
                img_uri = f"data:image/png;base64,{img_data.decode('utf-8')}"
                self.img_resources[img_key] = img_uri

        def lower(self) -> dict[str, Any]:
            """Pack data into a dictionary of the format that must be passed to
            `Visualizer.window.send`.
            """
            data = {
                "type": "set_object",
                "path": self.path.lower() if self.path is not None else "",
                "object": {
                    "metadata": {"version": 4.5, "type": "Object"},
                    "geometries": [],
                    "materials": [],
                    "object": {
                        "uuid": self.uuid,
                        "type": "_meshfile_object",
                        "format": "dae",
                        "data": self.dae_raw,
                        "resources": self.img_resources,
                        "matrix": list(self.intrinsic_transform.flatten()),
                    },
                },
            }
            if self.material is not None:
                self.material.lower_in_object(data)
            return data

        def set_scale(self, scale) -> None:
            self.intrinsic_transform[:3, :3] = np.diag(scale)

    # end code adapted from Jiminy

    class Plane(mg.Geometry):
        """A plane of the given width and height."""

        def __init__(
            self,
            width: float,
            height: float,
            widthSegments: float = 1,
            heightSegments: float = 1,
        ):
            super().__init__()
            self.width = width
            self.height = height
            self.widthSegments = widthSegments
            self.heightSegments = heightSegments

        def lower(self, object_data: Any) -> MsgType:
            return {
                "uuid": self.uuid,
                "type": "PlaneGeometry",
                "width": self.width,
                "height": self.height,
                "widthSegments": self.widthSegments,
                "heightSegments": self.heightSegments,
            }


if (
    WITH_HPP_FCL_BINDINGS
    and tuple(map(int, hppfcl.__version__.split("."))) >= (3, 0, 0)
    and hppfcl.WITH_OCTOMAP
):

    def loadOctree(octree: hppfcl.OcTree):
        boxes = octree.toBoxes()

        if len(boxes) == 0:
            return
        bs = boxes[0][3] / 2.0
        num_boxes = len(boxes)

        box_corners = np.array(
            [
                [bs, bs, bs],
                [bs, bs, -bs],
                [bs, -bs, bs],
                [bs, -bs, -bs],
                [-bs, bs, bs],
                [-bs, bs, -bs],
                [-bs, -bs, bs],
                [-bs, -bs, -bs],
            ]
        )

        all_points = np.empty((8 * num_boxes, 3))
        all_faces = np.empty((12 * num_boxes, 3), dtype=int)
        face_id = 0
        for box_id, box_properties in enumerate(boxes):
            box_center = box_properties[:3]

            corners = box_corners + box_center
            point_range = range(box_id * 8, (box_id + 1) * 8)
            all_points[point_range, :] = corners

            A = box_id * 8
            B = A + 1
            C = B + 1
            D = C + 1
            E = D + 1
            F = E + 1
            G = F + 1
            H = G + 1

            all_faces[face_id] = np.array([C, D, B])
            all_faces[face_id + 1] = np.array([B, A, C])
            all_faces[face_id + 2] = np.array([A, B, F])
            all_faces[face_id + 3] = np.array([F, E, A])
            all_faces[face_id + 4] = np.array([E, F, H])
            all_faces[face_id + 5] = np.array([H, G, E])
            all_faces[face_id + 6] = np.array([G, H, D])
            all_faces[face_id + 7] = np.array([D, C, G])
            # # top
            all_faces[face_id + 8] = np.array([A, E, G])
            all_faces[face_id + 9] = np.array([G, C, A])
            # # bottom
            all_faces[face_id + 10] = np.array([B, H, F])
            all_faces[face_id + 11] = np.array([H, B, D])

            face_id += 12

        colors = np.empty((all_points.shape[0], 3))
        colors[:] = np.ones(3)
        mesh = mg.TriangularMeshGeometry(all_points, all_faces, colors)
        return mesh

else:

    def loadOctree(octree):
        raise NotImplementedError("loadOctree need hppfcl with octomap support")


if WITH_HPP_FCL_BINDINGS:

    def loadMesh(mesh):
        if isinstance(mesh, (hppfcl.HeightFieldOBBRSS, hppfcl.HeightFieldAABB)):
            heights = mesh.getHeights()
            x_grid = mesh.getXGrid()
            y_grid = mesh.getYGrid()
            min_height = mesh.getMinHeight()

            X, Y = np.meshgrid(x_grid, y_grid)

            nx = len(x_grid) - 1
            ny = len(y_grid) - 1

            num_cells = (nx) * (ny) * 2 + (nx + ny) * 4 + 2

            num_vertices = X.size
            num_tris = num_cells

            faces = np.empty((num_tris, 3), dtype=int)
            vertices = np.vstack(
                (
                    np.stack(
                        (
                            X.reshape(num_vertices),
                            Y.reshape(num_vertices),
                            heights.reshape(num_vertices),
                        ),
                        axis=1,
                    ),
                    np.stack(
                        (
                            X.reshape(num_vertices),
                            Y.reshape(num_vertices),
                            np.full(num_vertices, min_height),
                        ),
                        axis=1,
                    ),
                )
            )

            face_id = 0
            for y_id in range(ny):
                for x_id in range(nx):
                    p0 = x_id + y_id * (nx + 1)
                    p1 = p0 + 1
                    p2 = p1 + nx + 1
                    p3 = p2 - 1

                    faces[face_id] = np.array([p0, p3, p1])
                    face_id += 1
                    faces[face_id] = np.array([p3, p2, p1])
                    face_id += 1

                    if y_id == 0:
                        p0_low = p0 + num_vertices
                        p1_low = p1 + num_vertices

                        faces[face_id] = np.array([p0, p1_low, p0_low])
                        face_id += 1
                        faces[face_id] = np.array([p0, p1, p1_low])
                        face_id += 1

                    if y_id == ny - 1:
                        p2_low = p2 + num_vertices
                        p3_low = p3 + num_vertices

                        faces[face_id] = np.array([p3, p3_low, p2_low])
                        face_id += 1
                        faces[face_id] = np.array([p3, p2_low, p2])
                        face_id += 1

                    if x_id == 0:
                        p0_low = p0 + num_vertices
                        p3_low = p3 + num_vertices

                        faces[face_id] = np.array([p0, p3_low, p3])
                        face_id += 1
                        faces[face_id] = np.array([p0, p0_low, p3_low])
                        face_id += 1

                    if x_id == nx - 1:
                        p1_low = p1 + num_vertices
                        p2_low = p2 + num_vertices

                        faces[face_id] = np.array([p1, p2_low, p2])
                        face_id += 1
                        faces[face_id] = np.array([p1, p1_low, p2_low])
                        face_id += 1

            # Last face
            p0 = num_vertices
            p1 = p0 + nx
            p2 = 2 * num_vertices - 1
            p3 = p2 - nx

            faces[face_id] = np.array([p0, p1, p2])
            face_id += 1
            faces[face_id] = np.array([p0, p2, p3])
            face_id += 1

        elif isinstance(mesh, (hppfcl.Convex, hppfcl.BVHModelBase)):
            if isinstance(mesh, hppfcl.BVHModelBase):
                num_vertices = mesh.num_vertices
                num_tris = mesh.num_tris

                call_triangles = mesh.tri_indices
                call_vertices = mesh.vertices

            elif isinstance(mesh, hppfcl.Convex):
                num_vertices = mesh.num_points
                num_tris = mesh.num_polygons

                call_triangles = mesh.polygons
                call_vertices = mesh.points

            faces = np.empty((num_tris, 3), dtype=int)
            for k in range(num_tris):
                tri = call_triangles(k)
                faces[k] = [tri[i] for i in range(3)]

            vertices = call_vertices()
            vertices = vertices.astype(np.float32)

        if num_tris > 0:
            mesh = mg.TriangularMeshGeometry(vertices, faces)
        else:
            mesh = mg.Points(
                mg.PointsGeometry(
                    vertices.T, color=np.repeat(np.ones((3, 1)), num_vertices, axis=1)
                ),
                mg.PointsMaterial(size=0.002),
            )

        return mesh

else:

    def loadMesh(mesh):
        raise NotImplementedError("loadMesh need hppfcl")


def loadPrimitive(geometry_object):
    import meshcat.geometry as mg

    # Cylinders need to be rotated
    R = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, -1.0, 0.0],
            [0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    RotatedCylinder = type(
        "RotatedCylinder", (mg.Cylinder,), {"intrinsic_transform": lambda self: R}
    )

    geom = geometry_object.geometry
    obj = None
    if WITH_HPP_FCL_BINDINGS and isinstance(geom, hppfcl.ShapeBase):
        if isinstance(geom, hppfcl.Capsule):
            if hasattr(mg, "TriangularMeshGeometry"):
                obj = createCapsule(2.0 * geom.halfLength, geom.radius)
            else:
                obj = RotatedCylinder(2.0 * geom.halfLength, geom.radius)
        elif isinstance(geom, hppfcl.Cylinder):
            obj = RotatedCylinder(2.0 * geom.halfLength, geom.radius)
        elif isinstance(geom, hppfcl.Cone):
            obj = RotatedCylinder(2.0 * geom.halfLength, 0, geom.radius, 0)
        elif isinstance(geom, hppfcl.Box):
            obj = mg.Box(npToTuple(2.0 * geom.halfSide))
        elif isinstance(geom, hppfcl.Sphere):
            obj = mg.Sphere(geom.radius)
        elif isinstance(geom, hppfcl.ConvexBase):
            obj = loadMesh(geom)

    if obj is None:
        msg = f"Unsupported geometry type for {geometry_object.name} ({type(geom)})"
        warnings.warn(msg, category=UserWarning, stacklevel=2)

    return obj


def createCapsule(length, radius, radial_resolution=30, cap_resolution=10):
    nbv = np.array([max(radial_resolution, 4), max(cap_resolution, 4)])
    h = length
    r = radius
    position = 0
    vertices = np.zeros((nbv[0] * (2 * nbv[1]) + 2, 3))
    for j in range(nbv[0]):
        phi = (2 * np.pi * j) / nbv[0]
        for i in range(nbv[1]):
            theta = (np.pi / 2 * i) / nbv[1]
            vertices[position + i, :] = np.array(
                [
                    np.cos(theta) * np.cos(phi) * r,
                    np.cos(theta) * np.sin(phi) * r,
                    -h / 2 - np.sin(theta) * r,
                ]
            )
            vertices[position + i + nbv[1], :] = np.array(
                [
                    np.cos(theta) * np.cos(phi) * r,
                    np.cos(theta) * np.sin(phi) * r,
                    h / 2 + np.sin(theta) * r,
                ]
            )
        position += nbv[1] * 2
    vertices[-2, :] = np.array([0, 0, -h / 2 - r])
    vertices[-1, :] = np.array([0, 0, h / 2 + r])
    indexes = np.zeros((nbv[0] * (4 * (nbv[1] - 1) + 4), 3))
    index = 0
    stride = nbv[1] * 2
    last = nbv[0] * (2 * nbv[1]) + 1
    for j in range(nbv[0]):
        j_next = (j + 1) % nbv[0]
        indexes[index + 0] = np.array(
            [
                j_next * stride + nbv[1],
                j_next * stride,
                j * stride,
            ]
        )
        indexes[index + 1] = np.array(
            [
                j * stride + nbv[1],
                j_next * stride + nbv[1],
                j * stride,
            ]
        )
        indexes[index + 2] = np.array(
            [
                j * stride + nbv[1] - 1,
                j_next * stride + nbv[1] - 1,
                last - 1,
            ]
        )
        indexes[index + 3] = np.array(
            [
                j_next * stride + 2 * nbv[1] - 1,
                j * stride + 2 * nbv[1] - 1,
                last,
            ]
        )
        for i in range(nbv[1] - 1):
            indexes[index + 4 + i * 4 + 0] = np.array(
                [
                    j_next * stride + i,
                    j_next * stride + i + 1,
                    j * stride + i,
                ]
            )
            indexes[index + 4 + i * 4 + 1] = np.array(
                [
                    j_next * stride + i + 1,
                    j * stride + i + 1,
                    j * stride + i,
                ]
            )
            indexes[index + 4 + i * 4 + 2] = np.array(
                [
                    j_next * stride + nbv[1] + i + 1,
                    j_next * stride + nbv[1] + i,
                    j * stride + nbv[1] + i,
                ]
            )
            indexes[index + 4 + i * 4 + 3] = np.array(
                [
                    j_next * stride + nbv[1] + i + 1,
                    j * stride + nbv[1] + i,
                    j * stride + nbv[1] + i + 1,
                ]
            )
        index += 4 * (nbv[1] - 1) + 4
    return mg.TriangularMeshGeometry(vertices, indexes)


class MeshcatVisualizer(BaseVisualizer):
    """A Pinocchio display using Meshcat"""

    FORCE_SCALE = 0.06
    FRAME_VEL_COLOR = 0x00FF00
    CAMERA_PRESETS: ClassVar = {
        "preset0": [
            np.zeros(3),  # target
            [3.0, 0.0, 1.0],  # anchor point (x, z, -y) lhs coords
        ],
        "preset1": [np.zeros(3), [1.0, 1.0, 1.0]],
        "preset2": [[0.0, 0.0, 0.6], [0.8, 1.0, 1.2]],
        "acrobot": [[0.0, 0.1, 0.0], [0.5, 0.0, 0.2]],
        "cam_ur": [[0.4, 0.6, -0.2], [1.0, 0.4, 1.2]],
        "cam_ur2": [[0.4, 0.3, 0.0], [0.5, 0.1, 1.4]],
        "cam_ur3": [[0.4, 0.3, 0.0], [0.6, 1.3, 0.3]],
        "cam_ur4": [[-1.0, 0.3, 0.0], [1.3, 0.1, 1.2]],  # x>0 to x<0
        "cam_ur5": [[-1.0, 0.3, 0.0], [-0.05, 1.5, 1.2]],
        "talos": [[0.0, 1.2, 0.0], [1.5, 0.3, 1.5]],
        "talos2": [[0.0, 1.1, 0.0], [1.2, 0.6, 1.5]],
    }

    def __init__(
        self,
        model=pin.Model(),
        collision_model=None,
        visual_model=None,
        copy_models=False,
        data=None,
        collision_data=None,
        visual_data=None,
    ):
        if not import_meshcat_succeed:
            msg = (
                "Error while importing the viewer client.\n"
                "Check whether meshcat is properly installed "
                "(pip install --user meshcat)."
            )
            raise ImportError(msg)

        super().__init__(
            model,
            collision_model,
            visual_model,
            copy_models,
            data,
            collision_data,
            visual_data,
        )
        self.static_objects = []

    def getViewerNodeName(self, geometry_object, geometry_type):
        """Return the name of the geometry object inside the viewer."""
        if geometry_type is pin.GeometryType.VISUAL:
            return self.viewerVisualGroupName + "/" + geometry_object.name
        elif geometry_type is pin.GeometryType.COLLISION:
            return self.viewerCollisionGroupName + "/" + geometry_object.name

    def initViewer(self, viewer=None, open=False, loadModel=False, zmq_url=None):
        """Start a new MeshCat server and client.
        Note: the server can also be started separately using the "meshcat-server"
        command in a terminal:
        this enables the server to remain active after the current script ends.
        """

        self.viewer = meshcat.Visualizer(zmq_url) if viewer is None else viewer

        self._node_default_cam = self.viewer["/Cameras/default"]
        self._node_background = self.viewer["/Background"]
        self._rot_cam_key = "rotated/<object>"
        self.static_objects = []

        self._check_meshcat_has_get_image()

        self._node_default_cam = self.viewer["/Cameras/default"]
        self._node_background = self.viewer["/Background"]
        self._rot_cam_key = "rotated/object"
        self.static_objects = []

        self._check_meshcat_has_get_image()

        if open:
            self.viewer.open()

        if loadModel:
            self.loadViewerModel()

    def reset(self):
        self.viewer.delete()
        self.static_objects = []

    def setBackgroundColor(self, preset_name: str = "gray", col_top=None, col_bot=None):
        """Set the background."""
        if col_top is not None:
            if col_bot is None:
                col_bot = col_top
        else:
            assert preset_name in COLOR_PRESETS.keys()
            col_top, col_bot = COLOR_PRESETS[preset_name]
        self._node_background.set_property("top_color", col_top)
        self._node_background.set_property("bottom_color", col_bot)

    def setCameraTarget(self, target: np.ndarray):
        self.viewer.set_cam_target(target)

    def setCameraPosition(self, position: np.ndarray):
        self.viewer.set_cam_pos(position)

    def setCameraPreset(self, preset_key: str):
        """Set the camera angle and position using a given preset."""
        assert preset_key in self.CAMERA_PRESETS
        cam_val = self.CAMERA_PRESETS[preset_key]
        self.setCameraTarget(cam_val[0])
        self.setCameraPosition(cam_val[1])

    def setCameraZoom(self, zoom: float):
        elt = self._node_default_cam[self._rot_cam_key]
        elt.set_property("zoom", zoom)

    def setCameraPose(self, pose):
        self._node_default_cam.set_transform(pose)

    def disableCameraControl(self):
        self.setCameraPosition([0, 0, 0])

    def enableCameraControl(self):
        self.setCameraPosition([3, 0, 1])

    def loadPrimitive(self, geometry_object: pin.GeometryObject):
        import meshcat.geometry as mg

        # Cylinders need to be rotated
        basic_three_js_transform = np.array(
            [
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, -1.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        )
        RotatedCylinder = type(
            "RotatedCylinder",
            (mg.Cylinder,),
            {"intrinsic_transform": lambda self: basic_three_js_transform},
        )

        # Cones need to be rotated

        geom = geometry_object.geometry
        obj = None
        if WITH_HPP_FCL_BINDINGS and isinstance(geom, hppfcl.ShapeBase):
            if isinstance(geom, hppfcl.Capsule):
                if hasattr(mg, "TriangularMeshGeometry"):
                    obj = createCapsule(2.0 * geom.halfLength, geom.radius)
                else:
                    obj = RotatedCylinder(2.0 * geom.halfLength, geom.radius)
            elif isinstance(geom, hppfcl.Cylinder):
                obj = RotatedCylinder(2.0 * geom.halfLength, geom.radius)
            elif isinstance(geom, hppfcl.Cone):
                obj = RotatedCylinder(2.0 * geom.halfLength, 0, geom.radius, 0)
            elif isinstance(geom, hppfcl.Box):
                obj = mg.Box(npToTuple(2.0 * geom.halfSide))
            elif isinstance(geom, hppfcl.Sphere):
                obj = mg.Sphere(geom.radius)
            elif isinstance(geom, hppfcl.Plane):
                To = np.eye(4)
                To[:3, 3] = geom.d * geom.n
                TranslatedPlane = type(
                    "TranslatedPlane",
                    (mg.Plane,),
                    {"intrinsic_transform": lambda self: To},
                )
                sx = geometry_object.meshScale[0] * 10
                sy = geometry_object.meshScale[1] * 10
                obj = TranslatedPlane(sx, sy)
            elif isinstance(geom, hppfcl.Ellipsoid):
                obj = mg.Ellipsoid(geom.radii)
            elif isinstance(geom, (hppfcl.Plane, hppfcl.Halfspace)):
                plane_transform: pin.SE3 = pin.SE3.Identity()
                # plane_transform.translation[:] = geom.d # Does not work
                plane_transform.rotation = pin.Quaternion.FromTwoVectors(
                    pin.ZAxis, geom.n
                ).toRotationMatrix()
                TransformedPlane = type(
                    "TransformedPlane",
                    (Plane,),
                    {"intrinsic_transform": lambda self: plane_transform.homogeneous},
                )
                obj = TransformedPlane(1000, 1000)
            elif isinstance(geom, hppfcl.ConvexBase):
                obj = loadMesh(geom)

        if obj is None:
            msg = f"Unsupported geometry type for {geometry_object.name} ({type(geom)})"
            warnings.warn(msg, category=UserWarning, stacklevel=2)
            obj = None

        return obj

    def loadMeshFromFile(self, geometry_object):
        # Mesh path is empty if Pinocchio is built without HPP-FCL bindings
        if geometry_object.meshPath == "":
            msg = (
                "Display of geometric primitives is supported only if "
                "pinocchio is build with HPP-FCL bindings."
            )
            warnings.warn(msg, category=UserWarning, stacklevel=2)
            return None

        # Get file type from filename extension.
        file_extension = Path(geometry_object.meshPath).suffix
        if file_extension.lower() == ".dae":
            obj = DaeMeshGeometry(geometry_object.meshPath)
        elif file_extension.lower() == ".obj":
            obj = mg.ObjMeshGeometry.from_file(geometry_object.meshPath)
        elif file_extension.lower() == ".stl":
            obj = mg.StlMeshGeometry.from_file(geometry_object.meshPath)
        else:
            msg = f"Unknown mesh file format: {geometry_object.meshPath}."
            warnings.warn(msg, category=UserWarning, stacklevel=2)
            obj = None

        return obj

    def loadViewerGeometryObject(self, geometry_object, geometry_type, color=None):
        """Load a single geometry object"""
        node_name = self.getViewerNodeName(geometry_object, geometry_type)
        meshcat_node = self.viewer[node_name]

        try:
            obj = None
            if WITH_HPP_FCL_BINDINGS:
                if isinstance(geometry_object.geometry, hppfcl.ShapeBase):
                    obj = self.loadPrimitive(geometry_object)
                elif (
                    tuple(map(int, hppfcl.__version__.split("."))) >= (3, 0, 0)
                    and hppfcl.WITH_OCTOMAP
                    and isinstance(geometry_object.geometry, hppfcl.OcTree)
                ):
                    obj = loadOctree(geometry_object.geometry)
                elif hasMeshFileInfo(geometry_object):
                    obj = self.loadMeshFromFile(geometry_object)
                elif isinstance(
                    geometry_object.geometry,
                    (
                        hppfcl.BVHModelBase,
                        hppfcl.HeightFieldOBBRSS,
                        hppfcl.HeightFieldAABB,
                    ),
                ):
                    obj = loadMesh(geometry_object.geometry)
            if obj is None and hasMeshFileInfo(geometry_object):
                obj = self.loadMeshFromFile(geometry_object)
            if obj is None:
                msg = (
                    "The geometry object named "
                    + geometry_object.name
                    + " is not supported by Pinocchio/MeshCat for vizualization."
                )
                warnings.warn(msg, category=UserWarning, stacklevel=2)
                return
        except Exception as e:
            msg = (
                "Error while loading geometry object: "
                f"{geometry_object.name}\nError message:\n{e}"
            )
            warnings.warn(msg, category=UserWarning, stacklevel=2)
            return

        if isinstance(obj, mg.Object):
            meshcat_node.set_object(obj)
        elif isinstance(obj, (mg.Geometry, mg.ReferenceSceneElement)):
            material = mg.MeshPhongMaterial()
            # Set material color from URDF, converting for triplet of doubles to a
            # single int.

            def to_material_color(rgba) -> int:
                """Convert rgba color as list into rgba color as int"""
                return (
                    int(rgba[0] * 255) * 256**2
                    + int(rgba[1] * 255) * 256
                    + int(rgba[2] * 255)
                )

            if color is None:
                meshColor = geometry_object.meshColor
            else:
                meshColor = color
            # Add transparency, if needed.
            material.color = to_material_color(meshColor)

            if float(meshColor[3]) != 1.0:
                material.transparent = True
                material.opacity = float(meshColor[3])

            geom_material = geometry_object.meshMaterial
            if geometry_object.overrideMaterial and isinstance(
                geom_material, pin.GeometryPhongMaterial
            ):
                material.emissive = to_material_color(geom_material.meshEmissionColor)
                material.specular = to_material_color(geom_material.meshSpecularColor)
                material.shininess = geom_material.meshShininess * 100.0

            if isinstance(obj, DaeMeshGeometry):
                obj.path = meshcat_node.path
                if geometry_object.overrideMaterial:
                    obj.material = material
                meshcat_node.window.send(obj)
            else:
                meshcat_node.set_object(obj, material)

    def loadViewerModel(
        self,
        rootNodeName="pinocchio",
        color=None,
        collision_color=None,
        visual_color=None,
    ):
        """Load the robot in a MeshCat viewer.
        Parameters:
            rootNodeName: name to give to the robot in the viewer
            color: deprecated and optional, color to give to both the collision
                and visual models of the robot. This setting overwrites any color
                specified in the robot description. Format is a list of four
                RGBA floating-point numbers (between 0 and 1)
            collision_color: optional, color to give to the collision model of
                the robot. Format is a list of four RGBA floating-point numbers
                (between 0 and 1)
            visual_color: optional, color to give to the visual model of
                the robot. Format is a list of four RGBA floating-point numbers
                (between 0 and 1)
        """
        if color is not None:
            warnings.warn(
                "The 'color' argument is deprecated and will be removed in a "
                "future version of Pinocchio. Consider using "
                "'collision_color' and 'visual_color' instead.",
                category=DeprecatedWarning,
            )
            collision_color = color
            visual_color = color

        # Set viewer to use to gepetto-gui.
        self.viewerRootNodeName = rootNodeName

        # Collisions
        self.viewerCollisionGroupName = self.viewerRootNodeName + "/" + "collisions"

        if self.collision_model is not None:
            for collision in self.collision_model.geometryObjects:
                self.loadViewerGeometryObject(
                    collision, pin.GeometryType.COLLISION, collision_color
                )
        self.displayCollisions(False)

        # Visuals
        self.viewerVisualGroupName = self.viewerRootNodeName + "/" + "visuals"
        if self.visual_model is not None:
            for visual in self.visual_model.geometryObjects:
                self.loadViewerGeometryObject(
                    visual, pin.GeometryType.VISUAL, visual_color
                )
        self.displayVisuals(True)

        # Frames
        self.viewerFramesGroupName = self.viewerRootNodeName + "/" + "frames"
        self.displayFrames(False)

    def reload(self, new_geometry_object, geometry_type=None):
        """Reload a geometry_object given by its name and its type"""
        if geometry_type == pin.GeometryType.VISUAL:
            geom_model = self.visual_model
        else:
            geom_model = self.collision_model
            geometry_type = pin.GeometryType.COLLISION

        geom_id = geom_model.getGeometryId(new_geometry_object.name)
        geom_model.geometryObjects[geom_id] = new_geometry_object

        self.delete(new_geometry_object, geometry_type)
        visual = geom_model.geometryObjects[geom_id]
        self.loadViewerGeometryObject(visual, geometry_type)

    def clean(self):
        self.viewer.delete()

    def delete(self, geometry_object, geometry_type):
        viewer_name = self.getViewerNodeName(geometry_object, geometry_type)
        self.viewer[viewer_name].delete()

    def display(self, q=None):
        """
        Display the robot at configuration q in the viewer by placing all the bodies
        """
        if q is not None:
            pin.forwardKinematics(self.model, self.data, q)

        if self.display_collisions:
            self.updatePlacements(pin.GeometryType.COLLISION)

        if self.display_visuals:
            self.updatePlacements(pin.GeometryType.VISUAL)

        if self.display_frames:
            self.updateFrames()

    def updatePlacements(self, geometry_type):
        if geometry_type == pin.GeometryType.VISUAL:
            geom_model = self.visual_model
            geom_data = self.visual_data
        else:
            geom_model = self.collision_model
            geom_data = self.collision_data

        pin.updateGeometryPlacements(self.model, self.data, geom_model, geom_data)
        for visual in geom_model.geometryObjects:
            visual_name = self.getViewerNodeName(visual, geometry_type)
            # Get mesh pose.
            M = geom_data.oMg[geom_model.getGeometryId(visual.name)]
            # Manage scaling: force scaling even if this should be normally handled by
            # MeshCat (but there is a bug here)
            geom = visual.geometry
            if WITH_HPP_FCL_BINDINGS and isinstance(
                geom, (hppfcl.Plane, hppfcl.Halfspace)
            ):
                T = M.copy()
                T.translation += M.rotation @ (geom.d * geom.n)
                T = T.homogeneous
            else:
                T = M.homogeneous

            # Update viewer configuration.
            T = applyScalingOnHomegeneousTransform(T, visual.meshScale)
            self.viewer[visual_name].set_transform(T)

        for visual in self.static_objects:
            visual_name = self.getViewerNodeName(visual, pin.GeometryType.VISUAL)
            M: pin.SE3 = visual.placement
            T = M.homogeneous
            T = applyScalingOnHomegeneousTransform(T, visual.meshScale)
            self.viewer[visual_name].set_transform(T)

    def addGeometryObject(self, obj: pin.GeometryObject, color=None):
        """Add a visual GeometryObject to the viewer, with an optional color."""
        self.loadViewerGeometryObject(obj, pin.GeometryType.VISUAL, color)
        self.static_objects.append(obj)

    def _check_meshcat_has_get_image(self):
        if not hasattr(self.viewer, "get_image"):
            warnings.warn(
                "meshcat.Visualizer does not have the get_image() method."
                " You need meshcat >= 0.2.0 to get this feature."
            )

    def captureImage(self, w=None, h=None):
        """Capture an image from the Meshcat viewer and return an RGB array."""
        if w is not None or h is not None:
            # pass arguments when either is not None
            img = self.viewer.get_image(w, h)
        else:
            img = self.viewer.get_image()
        img_arr = np.asarray(img)
        return img_arr

    def displayCollisions(self, visibility):
        """Set whether to display collision objects or not."""
        if self.collision_model is None:
            self.display_collisions = False
        else:
            self.display_collisions = visibility
        self.viewer[self.viewerCollisionGroupName].set_property("visible", visibility)

        if visibility:
            self.updatePlacements(pin.GeometryType.COLLISION)

    def displayVisuals(self, visibility):
        """Set whether to display visual objects or not."""
        if self.visual_model is None:
            self.display_visuals = False
        else:
            self.display_visuals = visibility
        self.viewer[self.viewerVisualGroupName].set_property("visible", visibility)

        if visibility:
            self.updatePlacements(pin.GeometryType.VISUAL)

    def displayFrames(self, visibility, frame_ids=None, axis_length=0.2, axis_width=2):
        """Set whether to display frames or not."""
        self.display_frames = visibility
        if visibility:
            self.initializeFrames(frame_ids, axis_length, axis_width)
        self.viewer[self.viewerFramesGroupName].set_property("visible", visibility)

    def initializeFrames(self, frame_ids=None, axis_length=0.2, axis_width=2):
        """Initializes the frame objects for display."""
        import meshcat.geometry as mg

        self.viewer[self.viewerFramesGroupName].delete()
        self.frame_ids = []

        for fid, frame in enumerate(self.model.frames):
            if frame_ids is None or fid in frame_ids:
                frame_viz_name = f"{self.viewerFramesGroupName}/{frame.name}"
                self.viewer[frame_viz_name].set_object(
                    mg.LineSegments(
                        mg.PointsGeometry(
                            position=axis_length * FRAME_AXIS_POSITIONS,
                            color=FRAME_AXIS_COLORS,
                        ),
                        mg.LineBasicMaterial(
                            linewidth=axis_width,
                            vertexColors=True,
                        ),
                    )
                )
                self.frame_ids.append(fid)

    def updateFrames(self):
        """
        Updates the frame visualizations with the latest transforms from model data.
        """
        pin.updateFramePlacements(self.model, self.data)
        for fid in self.frame_ids:
            frame_name = self.model.frames[fid].name
            frame_viz_name = f"{self.viewerFramesGroupName}/{frame_name}"
            self.viewer[frame_viz_name].set_transform(self.data.oMf[fid].homogeneous)

    def drawFrameVelocities(self, frame_id: int, v_scale=0.2, color=FRAME_VEL_COLOR):
        pin.updateFramePlacement(self.model, self.data, frame_id)
        vFr = pin.getFrameVelocity(
            self.model, self.data, frame_id, pin.LOCAL_WORLD_ALIGNED
        )
        line_group_name = f"ee_vel/{frame_id}"
        self._draw_vectors_from_frame(
            [v_scale * vFr.linear], [frame_id], [line_group_name], [color]
        )

    def _draw_vectors_from_frame(
        self,
        vecs: list[np.ndarray],
        frame_ids: list[int],
        vec_names: list[str],
        colors: list[int],
    ):
        """Draw vectors extending from given frames."""
        import meshcat.geometry as mg

        if len(vecs) != len(frame_ids) or len(vecs) != len(vec_names):
            return ValueError(
                "Number of vectors and frames IDs or names is inconsistent."
            )
        for i, (fid, v) in enumerate(zip(frame_ids, vecs)):
            frame_pos = self.data.oMf[fid].translation
            vertices = np.array([frame_pos, frame_pos + v]).astype(np.float32).T
            name = vec_names[i]
            geometry = mg.PointsGeometry(position=vertices)
            geom_object = mg.LineSegments(
                geometry, mg.LineBasicMaterial(color=colors[i])
            )
            prefix = self.viewerVisualGroupName + "/lines/" + name
            self.viewer[prefix].set_object(geom_object)


__all__ = ["MeshcatVisualizer"]



import numpy as np
import numpy.linalg as npl
from pinocchio.utils import eye, zero

"""
This file implements a sparse linear problem (quadric cost, linear constraints -- LCQP)
where the decision variables are denoted by x=(x1 ... xn), n being the number of
factors.
The problem can be written:
  min      Sum_i=1^p  || A_i x - b_i ||^2
x1...xn

so that    forall j=1:q   C_j x = d_i

Matrices A_i and C_j are block sparse, i.e. they are acting only on some (few) of the
variables x1 .. xn.

The file implements the main class FactorGraph, which stores the LCQP problem and solve
it.
It also provides a secondary class Factor, used to set up FactorGraph
"""


class Factor:
    """
    A factor is a part of a linear constraint corresponding either a cost ||A x - b|| or
    a constraint Cx = d.
    In both cases, we have Ax = sum A_i x_i, where some A_i are null. One object of
    class Factor stores one of the A_i, along with the correspond <i> index. It is
    simply a pair (index, matrix).

    This class is used as a arguments of some of the setup functions of FactorGraph.
    """

    def __init__(self, index, matrix):
        self.index = index
        self.matrix = matrix


class FactorGraph:
    """
    The class FactorGraph stores a block-sparse linear-constrained quadratic program
    (LCQP) of variable x=(x1...xn). The size of the problem is set up at construction of
    the object.
    Methods add_factor() and add_factor_constraint() are used to set up the problem.
    Method solve() is used to compute the solution to the problem.
    """

    def __init__(self, variableSize, nbVariables):
        """
        Initialize a QP sparse problem as min || A x - b || so that C x = d
        where  x = (x1, .., xn), and dim(xi) = variableSize and n = nbVariables
        After construction, A, b, C and d are allocated and set to 0.
        """
        self.nx = variableSize
        self.N = nbVariables
        self.A = zero([0, self.N * self.nx])
        self.b = zero(0)
        self.C = zero([0, self.N * self.nx])
        self.d = zero(0)

    def matrix_form_factor(self, factors):
        """
        Internal function: not designed to be called by the user.
        Create a factor matrix [ A1 0 A2 0 A3 ... ] where the Ai's are placed at
        the indexes of the factors.
        """
        assert len(factors) > 0
        nr = factors[0].matrix.shape[0]  # nb rows of the factor
        nc = self.nx * self.N  # nb cols

        # Define and fill the new rows to be added
        A = zero([nr, nc])  # new factor to be added to self.A
        for factor in factors:
            assert factor.matrix.shape == (nr, self.nx)
            A[:, self.nx * factor.index : self.nx * (factor.index + 1)] = factor.matrix
        return A

    def add_factor(self, factors, reference):
        """
        Add a factor || sum_{i} factor[i].matrix * x_{factor[i].index} - reference ||
        to the cost.
        """
        # Add the new rows to the cost matrix.
        self.A = np.vstack([self.A, self.matrix_form_factor(factors)])
        self.b = np.vstack([self.b, reference])

    def add_factor_constraint(self, factors, reference):
        """
        Add a factor sum_{i} factor[i].matrix * x_{factor[i].index} =  reference
        to the constraints.
        """
        # Add the new rows to the cost matrix.
        self.C = np.vstack([self.C, self.matrix_form_factor(factors)])
        self.d = np.vstack([self.d, reference])

    def solve(self, eps=1e-8):
        """
        Implement a LCQP solver, with numerical threshold eps.
        """
        Cp = npl.pinv(self.C, eps)
        xopt = Cp * self.d
        P = eye(self.nx * self.N) - Cp * self.C
        xopt += npl.pinv(self.A * P, eps) * (self.b - self.A * xopt)
        return xopt



import os
import unittest
from pathlib import Path

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


@unittest.skipUnless(pin.WITH_URDFDOM, "Needs URDFDOM")
class TestContactInverseDynamics(TestCase):
    def setUp(self):
        self.model_dir = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
        self.model_path = self.model_dir / "talos_data"
        self.urdf_filename = "talos_reduced.urdf"
        self.srdf_filename = "talos.srdf"
        self.urdf_model_path = self.model_path / "robots" / self.urdf_filename
        self.srdf_full_path = self.model_path / "srdf" / self.srdf_filename

    def load_model(self):
        self.model = pin.buildModelFromUrdf(
            self.urdf_model_path, pin.JointModelFreeFlyer()
        )
        pin.loadReferenceConfigurations(self.model, self.srdf_full_path)
        self.q0 = self.model.referenceConfigurations["half_sitting"]

    def test_call_to_contact_inverse_dynamics(self):
        self.load_model()
        model = self.model
        feet_name = ["left_sole_link", "right_sole_link"]
        frame_ids = [model.getFrameId(frame_name) for frame_name in feet_name]

        q = self.q0
        v = np.zeros(model.nv)
        a = np.zeros(model.nv)
        data = model.createData()

        contact_models_list = []
        contact_datas_list = []
        cones_list = []

        contact_models_vec = pin.StdVec_RigidConstraintModel()
        contact_datas_vec = pin.StdVec_RigidConstraintData()
        cones_vec = pin.StdVec_CoulombFrictionCone()

        for frame_id in frame_ids:
            frame = model.frames[frame_id]
            contact_model = pin.RigidConstraintModel(
                pin.ContactType.CONTACT_3D, model, frame.parentJoint, frame.placement
            )

            contact_models_list.append(contact_model)
            contact_datas_list.append(contact_model.createData())
            cones_list.append(pin.CoulombFrictionCone(0.4))

            contact_models_vec.append(contact_model)
            contact_datas_vec.append(contact_model.createData())
            cones_vec.append(pin.CoulombFrictionCone(0.4))

        constraint_dim = 0
        for m in contact_models_list:
            constraint_dim += m.size()

        dt = 1e-3
        R = np.zeros(constraint_dim)
        constraint_correction = np.zeros(constraint_dim)
        lambda_guess = np.zeros(constraint_dim)
        prox_settings = pin.ProximalSettings(1e-12, 1e-6, 1)
        # pin.initConstraintDynamics(model, data, contact_models_list)  # not needed

        # test 1 with vector of contact models, contact datas and cones
        tau1 = pin.contactInverseDynamics(
            model,
            data,
            q,
            v,
            a,
            dt,
            contact_models_vec,
            contact_datas_vec,
            cones_vec,
            R,
            constraint_correction,
            prox_settings,
            lambda_guess,
        )

        # test 2 with list of contact models, cones
        tau2 = pin.contactInverseDynamics(
            model,
            data,
            q,
            v,
            a,
            dt,
            contact_models_list,
            contact_datas_vec,
            cones_list,
            R,
            constraint_correction,
            prox_settings,
            lambda_guess,
        )

        # test 3 with list of contact models, contact datas and cones
        tau3 = pin.contactInverseDynamics(
            model,
            data,
            q,
            v,
            a,
            dt,
            contact_models_list,
            contact_datas_list,
            cones_list,
            R,
            constraint_correction,
            prox_settings,
            lambda_guess,
        )
        self.assertApprox(tau1, tau2)
        self.assertApprox(tau1, tau3)


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python3
# Plot benchmark results

import argparse
import itertools
import json
import typing
from pathlib import Path

import numpy as np
from bokeh.layouts import column
from bokeh.models import ColumnDataSource, Whisker
from bokeh.plotting import figure, show
from bokeh.transform import factor_cmap, jitter

TimeNS = typing.NewType("TimeNS", int)


class Variance:
    """Variance informations"""

    def __init__(
        self,
        mean: TimeNS,
        median: TimeNS,
        stddev: TimeNS,
    ):
        self.mean = mean
        self.median = median
        self.stddev = stddev

    @staticmethod
    def default():
        return Variance(TimeNS(0), TimeNS(0), TimeNS(0))

    def __repr__(self):
        return f"<Variance {self.mean=}, {self.median=}, {self.stddev=}>"


class Data:
    """Benchmarks data"""

    def __init__(self, name: str, times: list[TimeNS], variance: None | Variance):
        self.name = name
        self.times = times
        self.variance = variance

    @staticmethod
    def from_name(name: str):
        return Data(name, [], None)

    def __repr__(self):
        return f"<Data {self.name=}, {self.times=}, {self.variance=}>"


def make_time_ns(bench: dict) -> TimeNS:
    """Convert JSON output time into TimeNS"""
    t = float(bench["real_time"])
    if bench["time_unit"] == "ns":
        return TimeNS(int(t))
    elif bench["time_unit"] == "us":
        return TimeNS(int(t * 1e3))
    elif bench["time_unit"] == "ms":
        return TimeNS(int(t * 1e6))
    elif bench["time_unit"] == "s":
        return TimeNS(int(t * 1e9))


def parse_json_output(json_content: dict) -> list[Data]:
    """Parse google benchmark JSON output"""
    data_dict = {}
    variance_dict = {}
    for bench in json_content["benchmarks"]:
        run_name = bench["run_name"]
        run_type = bench["run_type"]
        if run_type == "aggregate":
            aggregate_name = bench["aggregate_name"]
            variance = variance_dict.setdefault(run_name, Variance.default())
            if aggregate_name == "mean":
                variance.mean = make_time_ns(bench)
            elif aggregate_name == "median":
                variance.median = make_time_ns(bench)
            elif aggregate_name == "stddev":
                variance.stddev = make_time_ns(bench)
        elif run_type == "iteration":
            data = data_dict.setdefault(run_name, Data.from_name(run_name))
            data.times.append(make_time_ns(bench))

    for k, v in variance_dict.items():
        data_dict[k].variance = v

    return list(data_dict.values())


def class_name(s: str) -> str:
    """Convert data name into a nice label"""
    return s.split("/")[1]


def compute_data_lower_upper(data: Data) -> tuple[float, float]:
    """Compute data lower and upper bounds to display Whisker box"""
    if data.variance:
        v = data.variance
        return (float(v.median - v.stddev), float(v.median + v.stddev))
    else:
        return (float("nan"), float("nan"))


def flatten_data(data: Data) -> list[tuple[str, TimeNS]]:
    """Return a list with data name and time"""
    return [(class_name(data.name), t) for t in data.times]


def plot_batch(datas: list[Data]):
    # Extract all batch classes
    classes = [class_name(d.name) for d in datas]

    # Compute lower and upper bounds
    lower_upper = np.array([compute_data_lower_upper(d) for d in datas])
    lower = lower_upper[:, 0]
    upper = lower_upper[:, 1]

    p = figure(
        height=768,
        sizing_mode="stretch_width",
        x_range=classes,
        background_fill_color="#efefef",
        title="Benchmark results",
    )
    p.xgrid.grid_line_color = None

    # Create whisker plot
    whisker_source = ColumnDataSource(data=dict(base=classes, upper=upper, lower=lower))
    error = Whisker(
        base="base",
        upper="upper",
        lower="lower",
        source=whisker_source,
        level="annotation",
        line_width=2,
    )
    error.upper_head.size = 20
    error.lower_head.size = 20
    p.add_layout(error)

    # Create scatter plot
    flat_data = np.array(
        list(itertools.chain.from_iterable([flatten_data(d) for d in datas]))
    )
    # We must convert back time to int to avoid rendering issues
    scatter_source = ColumnDataSource(
        data=dict(cl=flat_data[:, 0], time=flat_data[:, 1].astype(int))
    )
    # We use jitter to avoid plotting all class timing on the same X
    p.scatter(
        jitter("cl", 0.3, range=p.x_range),
        "time",
        source=scatter_source,
        alpha=0.5,
        size=13,
        line_color="white",
        color=factor_cmap("cl", "Light7", classes),
    )
    return p


def plot(datas: list[Data]):
    figures = []
    # Batch data by 7 because we use Light7 as color palette
    for data_batch in itertools.batched(datas, 7):
        figures.append(plot_batch(list(data_batch)))
    show(
        column(
            figures,
            sizing_mode="stretch_width",
        )
    )


def is_file(file: str) -> Path:
    p = Path(file)
    if not p.is_file():
        raise argparse.ArgumentTypeError(f"{file} is not a file")
    return p


def argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Plot benchmark results")
    parser.add_argument("json_output", help="Include directory", type=is_file)
    return parser


def main(args: list[str]):
    parser = argument_parser()
    args = parser.parse_args(args)

    datas = parse_json_output(json.loads(args.json_output.open().read()))
    plot(datas)


if __name__ == "__main__":
    import sys

    main(sys.argv[1:])



import unittest

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


class TestDeriavtives(TestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoidRandom()
        self.data = self.model.createData()

        qmax = np.full((self.model.nq, 1), np.pi)
        self.q = pin.randomConfiguration(self.model, -qmax, qmax)
        self.v = np.random.rand(self.model.nv)
        self.a = np.random.rand(self.model.nv)

        self.fext = []
        for _ in range(self.model.njoints):
            self.fext.append(pin.Force.Random())

    def test_rnea_derivatives(self):
        res = pin.computeRNEADerivatives(self.model, self.data, self.q, self.v, self.a)

        self.assertTrue(len(res) == 3)

        data2 = self.model.createData()
        pin.rnea(self.model, data2, self.q, self.v, self.a)

        self.assertApprox(self.data.ddq, data2.ddq)

        # With external forces
        res = pin.computeRNEADerivatives(
            self.model, self.data, self.q, self.v, self.a, self.fext
        )

        self.assertTrue(len(res) == 3)

        pin.rnea(self.model, data2, self.q, self.v, self.a, self.fext)

        self.assertApprox(self.data.ddq, data2.ddq)

    def test_generalized_gravity_derivatives(self):
        res = pin.computeGeneralizedGravityDerivatives(self.model, self.data, self.q)

        data2 = self.model.createData()
        ref, _, _ = pin.computeRNEADerivatives(
            self.model, data2, self.q, self.v * 0, self.a * 0
        )

        self.assertApprox(res, ref)

    def test_static_torque_derivatives(self):
        res = pin.computeStaticTorqueDerivatives(
            self.model, self.data, self.q, self.fext
        )

        data2 = self.model.createData()
        ref, _, _ = pin.computeRNEADerivatives(
            self.model, data2, self.q, self.v * 0, self.a * 0, self.fext
        )

        self.assertApprox(res, ref)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import pinocchio as pin


@unittest.skipUnless(pin.WITH_HPP_FCL, "Needs HPP-FCL")
class TestGeometryObjectBindings(unittest.TestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoid()
        self.collision_model = pin.buildSampleGeometryModelHumanoid(self.model)

    def test_name_get_set(self):
        col = self.collision_model.geometryObjects[0]
        self.assertTrue(col.name == "rleg_shoulder_object")
        col.name = "new_collision_name"
        self.assertTrue(col.name == "new_collision_name")

    def test_parent_get_set(self):
        col = self.collision_model.geometryObjects[0]
        self.assertTrue(col.parentJoint == 2)
        col.parentJoint = 3
        self.assertTrue(col.parentJoint == 3)

    def test_placement_get_set(self):
        m = pin.SE3.Identity()
        new_m = pin.SE3.Random()
        col = self.collision_model.geometryObjects[0]
        self.assertTrue(np.allclose(col.placement.homogeneous, m.homogeneous))
        col.placement = new_m
        self.assertTrue(np.allclose(col.placement.homogeneous, new_m.homogeneous))

    def test_meshpath_get(self):
        col = self.collision_model.geometryObjects[0]
        self.assertTrue(col.meshPath is not None)

    def test_create_data(self):
        collision_data = self.collision_model.createData()
        self.assertEqual(len(collision_data.oMg), self.collision_model.ngeoms)

    def test_create_datas(self):
        collision_data = self.collision_model.createData()

        self.assertEqual(len(collision_data.oMg), self.collision_model.ngeoms)

        data_2, collision_data_2 = pin.createDatas(self.model, self.collision_model)
        self.assertTrue(self.model.check(data_2))
        self.assertEqual(len(collision_data_2.oMg), self.collision_model.ngeoms)

    def test_copy(self):
        collision_model_copy = self.collision_model.copy()
        self.assertEqual(self.collision_model.ngeoms, collision_model_copy.ngeoms)

        collision_data = self.collision_model.createData()
        collision_data_copy = collision_data.copy()
        self.assertEqual(len(collision_data.oMg), len(collision_data_copy.oMg))

    def test_material_get_set(self):
        col = self.collision_model.geometryObjects[0]
        self.assertTrue(isinstance(col.meshMaterial, (pin.GeometryNoMaterial)))
        col.meshMaterial = pin.GeometryPhongMaterial()
        self.assertTrue(isinstance(col.meshMaterial, (pin.GeometryPhongMaterial)))
        material = col.meshMaterial
        self.assertTrue(isinstance(material, (pin.GeometryPhongMaterial)))
        material.meshEmissionColor = np.array([1.0, 1.0, 1.0, 1.0])
        material.meshShininess = 0.7
        self.assertTrue(
            (material.meshEmissionColor == col.meshMaterial.meshEmissionColor).all()
        )
        self.assertEqual(material.meshShininess, col.meshMaterial.meshShininess)


if __name__ == "__main__":
    unittest.main()



# NOTE: this example needs gepetto-gui to be installed
# usage: launch gepetto-gui and then run this test

import os
import sys
from pathlib import Path

import pinocchio as pin
from pinocchio.visualize import GepettoVisualizer

# Load the URDF model.
model_path = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
mesh_dir = model_path.parent.parent
urdf_filename = "talos_reduced.urdf"
urdf_model_path = model_path / "talos_data/robots" / urdf_filename

model, collision_model, visual_model = pin.buildModelsFromUrdf(
    urdf_model_path, mesh_dir, pin.JointModelFreeFlyer()
)
viz = GepettoVisualizer(model, collision_model, visual_model)

# Initialize the viewer.
try:
    viz.initViewer()
except ImportError as err:
    print(
        "Error while initializing the viewer. "
        "It seems you should install gepetto-viewer"
    )
    print(err)
    sys.exit(0)

try:
    viz.loadViewerModel("pinocchio")
except AttributeError as err:
    print(
        "Error while loading the viewer model. It seems you should start gepetto-viewer"
    )
    print(err)
    sys.exit(0)

# Display a robot configuration.
q0 = pin.neutral(model)
viz.display(q0)

# Display another robot.
viz2 = GepettoVisualizer(model, collision_model, visual_model)
viz2.initViewer(viz.viewer)
viz2.loadViewerModel(rootNodeName="pinocchio2")
q = q0.copy()
q[1] = 1.0
viz2.display(q)



import unittest

import numpy as np
import pinocchio as pin
from pinocchio.utils import eye, rand, zero
from test_case import PinocchioTestCase as TestCase


class TestInertiaBindings(TestCase):
    def test_zero_getters(self):
        Y = pin.Inertia.Zero()
        self.assertTrue(Y.mass == 0)
        self.assertTrue(np.allclose(zero(3), Y.lever))
        self.assertTrue(np.allclose(zero([3, 3]), Y.inertia))

    def test_identity_getters(self):
        Y = pin.Inertia.Identity()
        self.assertTrue(Y.mass == 1)
        self.assertTrue(np.allclose(zero(3), Y.lever))
        self.assertTrue(np.allclose(eye(3), Y.inertia))

    # TODO: this is not nice, since a random matrix can *in theory* be unity
    def test_setRandom(self):
        Y = pin.Inertia.Identity()
        Y.setRandom()
        self.assertFalse(Y.mass == 1)
        self.assertFalse(np.allclose(zero(3), Y.lever))
        self.assertFalse(np.allclose(eye(3), Y.inertia))

    def test_setZero(self):
        Y = pin.Inertia.Zero()
        Y.setRandom()
        Y.setZero()
        self.assertTrue(Y.mass == 0)
        self.assertTrue(np.allclose(zero(3), Y.lever))
        self.assertTrue(np.allclose(zero([3, 3]), Y.inertia))

    def test_setIdentity(self):
        Y = pin.Inertia.Zero()
        Y.setIdentity()
        self.assertTrue(Y.mass == 1)
        self.assertTrue(np.allclose(zero(3), Y.lever))
        self.assertTrue(np.allclose(eye(3), Y.inertia))

    def test_set_mass(self):
        Y = pin.Inertia.Zero()
        Y.mass = 10
        self.assertTrue(np.allclose(Y.mass, 10))

    def test_set_lever(self):
        Y = pin.Inertia.Zero()
        lev = rand(3)
        Y.lever = lev
        self.assertTrue(np.allclose(Y.lever, lev))

    def test_set_inertia(self):
        Y = pin.Inertia.Zero()
        iner = rand([3, 3])
        iner = (iner + iner.T) / 2.0  # Symmetrize the matrix
        Y.inertia = iner
        self.assertTrue(np.allclose(Y.inertia, iner))

    def test_internal_sums(self):
        Y1 = pin.Inertia.Random()
        Y2 = pin.Inertia.Random()
        Y = Y1 + Y2
        self.assertTrue(np.allclose(Y1.matrix() + Y2.matrix(), Y.matrix()))

    def test_se3_action(self):
        m = pin.SE3.Random()
        Y = pin.Inertia.Random()
        v = pin.Motion.Random()
        self.assertTrue(np.allclose((Y * v).vector, Y.matrix().dot(v.vector)))
        self.assertTrue(
            np.allclose(
                (m * Y).matrix(),
                m.inverse().action.T.dot(Y.matrix()).dot(m.inverse().action),
            )
        )
        self.assertTrue(
            np.allclose(
                m.act(Y).matrix(),
                m.inverse().action.T.dot(Y.matrix()).dot(m.inverse().action),
            )
        )
        self.assertTrue(
            np.allclose(
                (m.actInv(Y)).matrix(), m.action.T.dot(Y.matrix()).dot(m.action)
            )
        )

    def test_dynamic_parameters(self):
        In = pin.Inertia.Random()

        v = In.toDynamicParameters()

        self.assertApprox(v[0], In.mass)
        self.assertApprox(v[1:4], In.mass * In.lever)

        I_o = In.inertia + In.mass * pin.skew(In.lever).transpose().dot(
            pin.skew(In.lever)
        )
        I_ov = np.array(
            [
                [float(v[4]), float(v[5]), float(v[7])],
                [float(v[5]), float(v[6]), float(v[8])],
                [float(v[7]), float(v[8]), float(v[9])],
            ]
        )

        self.assertApprox(I_o, I_ov)

        I2 = pin.Inertia.FromDynamicParameters(v)
        self.assertApprox(I2, In)

        # Test FromDynamicParameters raise an exception if the wrong
        # vector size is provided
        with self.assertRaises(ValueError):
            pin.Inertia.FromDynamicParameters(np.array([]))

    def test_pseudo_inertia(self):
        In = pin.Inertia.Random()

        pseudo = In.toPseudoInertia()

        # test accessing mass, h, sigma
        self.assertApprox(pseudo.mass, In.mass)
        self.assertApprox(pseudo.h, In.mass * In.lever)
        self.assertEqual(pseudo.sigma.shape, (3, 3))

        # test toMatrix
        _ = pseudo.toMatrix()

        # test toDynamicParameters
        params = pseudo.toDynamicParameters()

        # test fromDynamicParameters
        pseudo2 = pin.PseudoInertia.FromDynamicParameters(params)
        self.assertApprox(pseudo.mass, pseudo2.mass)
        self.assertApprox(pseudo.h, pseudo2.h)
        self.assertApprox(pseudo.sigma, pseudo2.sigma)

        # Test FromDynamicParameters raise an exception if the wrong
        # vector size is provided
        with self.assertRaises(ValueError):
            pin.PseudoInertia.FromDynamicParameters(np.array([]))

        # test fromMatrix
        pseudo3 = pin.PseudoInertia.FromMatrix(pseudo.toMatrix())
        self.assertApprox(pseudo.mass, pseudo3.mass)
        self.assertApprox(pseudo.h, pseudo3.h)
        self.assertApprox(pseudo.sigma, pseudo3.sigma)

        # test fromInertia
        pseudo4 = pin.PseudoInertia.FromInertia(In)
        self.assertApprox(pseudo.mass, pseudo4.mass)
        self.assertApprox(pseudo.h, pseudo4.h)
        self.assertApprox(pseudo.sigma, pseudo4.sigma)

        # test toInertia
        self.assertApprox(pseudo4.toInertia(), In)

        # test from PseudoInertia
        pin.PseudoInertia(pseudo)

        # test array
        pseudo_array = np.array(pseudo)
        self.assertApprox(pseudo_array, pseudo.toMatrix())

    def test_log_cholesky(self):
        log_cholesky = pin.LogCholeskyParameters(np.random.randn(10))
        In = pin.Inertia.FromLogCholeskyParameters(log_cholesky)

        # Test constructor with wrong vector size
        with self.assertRaises(ValueError):
            log_cholesky = pin.LogCholeskyParameters(np.array([]))

        # test accessing parameters
        self.assertEqual(log_cholesky.parameters.shape, (10,))

        # test toDynamicParameters
        params = log_cholesky.toDynamicParameters()
        params2 = In.toDynamicParameters()
        self.assertApprox(params, params2)

        # test toPseudoInertia
        pseudo = log_cholesky.toPseudoInertia()
        pseudo2 = In.toPseudoInertia()
        self.assertApprox(pseudo.mass, pseudo2.mass)
        self.assertApprox(pseudo.h, pseudo2.h)
        self.assertApprox(pseudo.sigma, pseudo2.sigma)

        # test toInertia
        In2 = log_cholesky.toInertia()
        self.assertApprox(In, In2)

        # test calculateJacobian
        log_cholesky.calculateJacobian()

        # test array
        log_cholesky_array = np.array(log_cholesky)
        self.assertApprox(log_cholesky_array, log_cholesky.parameters)

    def test_array(self):
        In = pin.Inertia.Random()
        I_array = np.array(In)

        self.assertApprox(I_array, In.matrix())

    def test_several_init(self):
        for _ in range(100000):
            In = pin.Inertia.Random() + pin.Inertia.Random()
            s = In.__str__()
            self.assertTrue(s != "")


if __name__ == "__main__":
    unittest.main()
